"""
From https://github.com/shish/sqliteshelf (MIT License)

by default, things are stored in a "shelf" table

>>> d = SQLiteShelf("test.sdb")

you can put multiple shelves into a single SQLite database

>>> e = SQLiteShelf("test.sdb", "othertable")

both are empty to start with

>>> d
{}
>>> e
{}

adding stuff is as simple as a regular dict
>>> d['a'] = "moo"
>>> e['a'] = "moo"

regular dict actions work

>>> d['a']
'moo'
>>> e['a']
'moo'
>>> 'a' in d
True
>>> len(d)
1
>>> del d['a']
>>> 'a' in d
False
>>> len(d)
0
>>> del e['a']

objects can be stored in shelves

>> class Test:
..    def __init__(self):
..        self.foo = "bar"
..
>> t = Test()
>> d['t'] = t
>> print d['t'].foo
bar

errors are as normal for a dict

>>> d['x']
Traceback (most recent call last):
    ...
KeyError: 'x'
>>> del d['x']
Traceback (most recent call last):
    ...
KeyError: 'x'

Adding and fetching binary strings

>>> d[1] = "a\\x00b"
>>> d[1]
'a\\x00b'
"""

try:
    from UserDict import DictMixin
except ImportError:
    from collections.abc import MutableMapping as DictMixin

try:
    import cPickle as pickle
except ImportError:
    import pickle

import sys

valtype = str
if sys.version > "3":
    buffer = memoryview
    valtype = bytes

import sqlite3  # noqa: E402


class SQLiteDict(DictMixin):
    def __init__(
        self, filename=":memory:", table="shelf", flags="r", mode=None, valtype=valtype
    ):
        self.table = table
        self.valtype = valtype
        MAKE_SHELF = (
            "CREATE TABLE IF NOT EXISTS " + self.table + " (key TEXT, value BLOB)"
        )
        MAKE_INDEX = (
            "CREATE UNIQUE INDEX IF NOT EXISTS "
            + self.table
            + "_keyndx ON "
            + self.table
            + "(key)"
        )
        self.conn = sqlite3.connect(filename)
        self.conn.text_factory = str
        self.conn.execute(MAKE_SHELF)
        self.conn.execute(MAKE_INDEX)
        self.conn.commit()

    def __getitem__(self, key):
        GET_ITEM = "SELECT value FROM " + self.table + " WHERE key = ?"
        item = self.conn.execute(GET_ITEM, (key,)).fetchone()
        if item is None:
            raise KeyError(key)
        return self.valtype(item[0])

    def __setitem__(self, key, item):
        ADD_ITEM = "REPLACE INTO " + self.table + " (key, value) VALUES (?,?)"
        self.conn.execute(ADD_ITEM, (key, sqlite3.Binary(item)))
        self.conn.commit()

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        DEL_ITEM = "DELETE FROM " + self.table + " WHERE key = ?"
        self.conn.execute(DEL_ITEM, (key,))
        self.conn.commit()

    def __iter__(self):
        c = self.conn.cursor()
        try:
            c.execute("SELECT key FROM " + self.table + " ORDER BY key")
            for row in c:
                yield row[0]
        finally:
            c.close()

    def keys(self):
        c = self.conn.cursor()
        try:
            c.execute("SELECT key FROM " + self.table + " ORDER BY key")
            return [row[0] for row in c]
        finally:
            c.close()

    ###################################################################
    # optional bits

    def __len__(self):
        GET_LEN = "SELECT COUNT(*) FROM " + self.table
        return self.conn.execute(GET_LEN).fetchone()[0]

    def close(self):
        if self.conn is not None:
            self.conn.commit()
            self.conn.close()
            self.conn = None

    def __del__(self):
        self.close()

    def __repr__(self):
        return repr(dict(self))


class SQLiteShelf(SQLiteDict):
    def __getitem__(self, key):
        return pickle.loads(SQLiteDict.__getitem__(self, key))

    def __setitem__(self, key, item):
        SQLiteDict.__setitem__(self, key, pickle.dumps(item))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#! /usr/bin/env python
import asyncio
import json

from pathlib import Path

import defopt

from structlog.stdlib import get_logger

from ice import execution_context
from ice.cli_utils import select_recipe_class
from ice.environment import env
from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import retrieve_gold_standards_df
from ice.mode import Mode
from ice.paper import Paper
from ice.recipe import is_list_of_recipe_result
from ice.recipe import Recipe
from ice.trace import enable_trace
from ice.trace import trace
from ice.utils import map_async


log = get_logger()


def main_cli(
    *,
    mode: Mode = "machine",
    output_file: str | None = None,
    json_out: str | None = None,
    recipe_name: str | None = None,
    input_files: list[str] | None = None,
    gold_standard_splits: list[str] | None = None,
    question_short_name: str | None = None,
    trace: bool = True,
    args: dict | None = None,
):
    """
    ::
    Run a recipe.
    :param mode Mode:
    :param output_file:  Append output to a file in markdown format instead of stdout.
    :param json_out:  Write recipe-specific JSON output to a file.
    :param recipe_name:  Name of the recipe to run.
    :param input_files:  List of files to run recipe over.
    :param gold_standard_splits: "iterate", "validation", and/or "test"
    """
    if trace:
        enable_trace()

    async def main_wrapper():
        # A traced function cannot be called until the event loop is running.
        return await main(
            mode=mode,
            output_file=output_file,
            json_out=json_out,
            recipe_name=recipe_name,
            input_files=input_files,
            gold_standard_splits=gold_standard_splits,
            question_short_name=question_short_name,
            args=args or {},
        )

    asyncio.run(main_wrapper())


@trace
async def main(
    *,
    mode: Mode,
    output_file: str | None,
    json_out: str | None,
    recipe_name: str | None,
    input_files: list[str] | None,
    gold_standard_splits: list[str] | None,
    question_short_name: str | None,
    args: dict,
):
    # User selects recipe
    recipe = await get_recipe(recipe_name, mode)

    # User selects papers
    papers = await get_papers(input_files, gold_standard_splits, question_short_name)

    if papers:
        print(
            f"Running recipe {recipe} over papers {', '.join(p.document_id for p in papers)}"
        )

    # Run recipe without paper arguments
    if not papers:
        result = await recipe.run(**args)
        env().print(
            result,
            format_markdown=False,
            file=output_file,
        )
        return

    # Run recipe over papers
    results_by_doc = await run_recipe_over_papers(recipe, papers, args)

    # Print results
    results_json = await print_results(recipe, results_by_doc, output_file, json_out)

    # Print evaluation of results
    await evaluate_results(recipe, results_json, output_file)


async def get_recipe(recipe_name: str | None, mode: Mode) -> Recipe:
    """
    Get the recipe instance based on the user input or selection.
    """
    recipe_class = await select_recipe_class(recipe_name=recipe_name)
    return recipe_class(mode)


async def get_papers(
    input_files: list[str] | None,
    gold_standard_splits: list[str] | None,
    question_short_name: str | None,
) -> list[Paper]:
    """
    Get the list of papers based on the user input or selection.
    """
    if (gold_standard_splits is None) != (question_short_name is None):
        raise ValueError(
            "Must specify both gold_standard_splits and question_short_name or neither."
        )

    if input_files:
        paper_files = [Path(i) for i in input_files]
    elif gold_standard_splits:
        gs_df = retrieve_gold_standards_df()
        question_gs_in_splits = gs_df[
            (gs_df.question_short_name == question_short_name)
            & (gs_df.split.isin(gold_standard_splits))
            & (gs_df["Are quotes enough?"] != "No")
        ]
        paper_dir = Path(__file__).parent / "papers/"
        paper_files = [
            f
            for f in paper_dir.iterdir()
            if f.name in question_gs_in_splits.document_id.unique()
        ]
    else:
        paper_files = []

    # If user doesn't specify papers via CLI args, we could prompt them
    # but this makes it harder to run recipes that don't take papers as
    # arguments, so we won't do that here.

    # if input_files is None and gold_standard_splits is None:
    #     paper_names = [f.name for f in paper_files]
    #     selected_paper_names = await env().checkboxes("Papers", paper_names)
    #     paper_files = [f for f in paper_files if f.name in selected_paper_names]

    return [Paper.load(f) for f in paper_files]


async def run_recipe_over_papers(
    recipe: Recipe, papers: list[Paper], args: dict
) -> dict[str, RecipeResult]:
    """
    Run the recipe over the papers and return a map from paper ids to recipe results.
    """

    async def apply_recipe_to_paper(paper: Paper):
        execution_context.new_context(document_id=paper.document_id, task=str(recipe))
        return await recipe.run(paper=paper, **args)

    # Run recipe over papers
    max_concurrency = 5 if recipe.mode == "machine" else 1
    results = await map_async(
        papers,
        apply_recipe_to_paper,
        show_progress_bar=True,
        max_concurrency=max_concurrency,
    )

    return {paper.document_id: result for (paper, result) in zip(papers, results)}


async def print_results(
    recipe: Recipe,
    results_by_doc: dict[str, RecipeResult],
    output_file: str | None,
    json_out: str | None,
) -> list[dict]:
    """
    Print the results to the output file or stdout, and return the JSON representation of the results.
    """
    results_json: list[dict] = []

    for (document_id, final_result) in results_by_doc.items():

        if json_out is not None:
            results_json.extend(recipe.to_json(final_result))

        env().print(
            f"## Final result for {document_id}\n",
            format_markdown=False if output_file else True,
            wait_for_confirmation=False,
            file=output_file,
        )

        if is_list_of_recipe_result(final_result):
            results_to_print = [r.result for r in final_result]
        else:
            results_to_print = [final_result]

        for result_to_print in results_to_print:
            env().print(
                result_to_print,
                format_markdown=False,
                wait_for_confirmation=False,
                file=output_file,
            )

    if json_out is not None:
        with open(json_out, "w") as f:
            json.dump(results_json, f, indent=2)

    return results_json


async def evaluate_results(
    recipe: Recipe, results_json: list[dict], output_file: str | None
):
    """
    Evaluate the results using the recipe's evaluation report and
    dashboard row methods, and print the report to the output file or
    stdout.
    """
    if recipe.results:
        evaluation_report = await recipe.evaluation_report()

        env().print(
            evaluation_report,
            format_markdown=False if output_file else True,
            wait_for_confirmation=True,
            file=output_file,
        )

        evaluation_report.make_dashboard_row_df()
        evaluation_report.make_experiments_evaluation_df()


if __name__ == "__main__":
    defopt.run(main_cli, parsers={dict: json.loads})



from ice.agents.base import Agent
from ice.recipe import recipe
from ice.recipes.primer.debate.prompt import *


async def turn(debate: Debate, agent: Agent, agent_name: Name, turns_left: int):
    prompt = render_debate_prompt(agent_name, debate, turns_left)
    answer = await agent.complete(prompt=prompt, stop="\n")
    return (agent_name, answer.strip('" '))


async def debate(question: str = "Should we legalize all drugs?"):
    agents = [recipe.agent(), recipe.agent()]
    agent_names = ["Alice", "Bob"]
    debate = initialize_debate(question)
    turns_left = 8
    while turns_left > 0:
        for agent, agent_name in zip(agents, agent_names):
            response = await turn(debate, agent, agent_name, turns_left)
            debate.append(response)
            turns_left -= 1
    return render_debate(debate)


recipe.main(debate)



from transformers import GPT2TokenizerFast

from ice.recipe import recipe
from ice.recipes.abstract_qa import Abstract
from ice.recipes.abstract_qa import DEFAULT_ABSTRACTS


def make_gpt2_tokenizer() -> GPT2TokenizerFast:
    return GPT2TokenizerFast.from_pretrained("gpt2")


gpt2_tokenizer = make_gpt2_tokenizer()


def num_tokens(text: str) -> int:
    """
    Return how many tokens are in 'text'.
    """
    return len(gpt2_tokenizer.tokenize(text))


PREFIX = """In this section, we will demonstrate how to write an ideal answer for a question using academic literature. When answering questions using academic literature you MUST use references.
An ideal answer gives references to the academic literature. Example: "To our knowledge, the only freely and publicly available dense autoregressive language models larger than GPT2 are GPT-Neo (Black et al., 2021), GPT-J-6B (Wang and Komatsuzaki, 2021), Megatron-11B, Pangu-13B (Zeng et al., 2021), and the recently released FairSeq models (Artetxe et al., 2021)."
Let's look at some example questions."""

MAX_TOKENS = 4001

MIN_COMPLETION_TOKENS = 250

SHOTS = [
    """Question: "How does l-theanine affect anxiety?"

Here are the relevant papers, and excerpts from those papers:

Paper: Effects of l-theanine on anxiety-like behavior, cerebrospinal fluid amino acid profile, and hippocampal activity in Wistar Kyoto rats
Reference: Ogawa et al. (2017)
Excerpt: Rationale and objectives In the present study, we examined the effects of repeated l-theanine administration on behavior, levels of amino acids in the cerebrospinal fluid (CSF), and hippocampal activity in Wistar Kyoto (WKY) rats, an animal model of anxiety and depressive disorders.MethodsBehavioral tests were performed after 7–10 days of l-theanine (0.4 mg kg−1 day−1) or saline administration, followed by CSF sampling for high-performance liquid chromatography (HPLC) analysis. An independent set of animals was subjected to [18F]fluorodeoxyglucose positron emission tomography (PET) scanning after the same dose of l-theanine or saline administration for 7 days.ResultsIn the elevated plus maze test, the time spent in the open arms was significantly longer in the l-theanine group than in the saline group (P = 0.035). In addition, significantly lower CSF glutamate (P = 0.039) and higher methionine (P = 0.024) concentrations were observed in the l-theanine group than in the saline group. A significant increase in the standard uptake value ratio was observed in the hippocampus/cerebellum of the l-theanine group (P < 0.001). ConclusionsThese results suggest that l-theanine enhances hippocampal activity and exerts anxiolytic effects, which may be mediated by changes in glutamate and methionine levels in the brain.

Paper: L-theanine—a unique amino acid of green tea and its relaxation effect in humans
Reference: Raj Juneja et al. (1999)
Excerpt: It was found that L-theanine administered intraperitoneally to rats reached the brain within 30 min without any metabolic change. Theanine also acts as a neurotransmitter in the brain and decreased blood pressure significantly in hypertensive rats. In general, animals always generate very weak electric pulses on the surface of the brain, called brain waves. Brain waves are classified into four types, namely α,β,δ and θ-waves, based on mental conditions. Generation of α-waves is considered to be an index of relaxation. In human volunteers, α-waves were generated on the occipital and parietal regions of the brain surface within 40 min after the oral administration of theanine (50–200 mg), signifying relaxation without causing drowsiness.

Paper: The acute effects of L-theanine in comparison with alprazolam on anticipatory anxiety in humans
Reference: Lu et al. (2004)
Excerpt: Sixteen healthy volunteers received alprazolam (1 mg), L‐theanine (200 mg) or placebo in a double‐blind placebo‐controlled repeated measures design. The acute effects of alprazolam and L‐theanine were assessed under a relaxed and experimentally induced anxiety condition. Subjective self‐reports of anxiety including BAI, VAMS, STAI state anxiety, were obtained during both task conditions at pre‐ and post‐drug administrations. The results showed some evidence for relaxing effects of L‐theanine during the baseline condition on the tranquil–troubled subscale of the VAMS. Alprazolam did not exert any anxiolytic effects in comparison with the placebo on any of the measures during the relaxed state. Neither L‐theanine nor alprazalam had any significant anxiolytic effects during the experimentally induced anxiety state. The findings suggest that while L‐theanine may have some relaxing effects under resting conditions, neither L‐theanine not alprazolam demonstrate any acute anxiolytic effects under conditions of increased anxiety in the AA model.

Ideal answer to the question "How does l-theanine affect anxiety?" based on the papers above using references: Two of the three papers indicate that l-theanine reduces anxiety in animals (Ogawa et al., 2017) and humans (Kimura et al., 2007), while one study indicates that l-theanine may not be effective for anxiety (Lu et al., 2004).""",
    """Question: "How many emotions are there?"

Here are the relevant papers, and excerpts from those papers:

Paper: How Many Emotions Are There? Wedding the Social and the Autonomic Components
Reference: Kemper (1987)
Excerpt: Fundamental in the field of emotions is the question of how many emotion there are or there can be. The answer proposed here is that the number of possible emotions is limites. As long as society differentiates new social situations, labels them, and socializes individuals to experience them, new emotions will continue to emerge. But this view must be qualified by an understanding of the autonomic constraints that limit variability in the experience of emotions. It is argued here that there are four psychologically grounded primary emotions: fear, anger, depression, and satisfaction. They are evolutionarily important, cross-culturally universal, ontogenetically early to emerge, and link empirically with important outcomes of social relations. Secondary emotions, such as guilt, shame, pride, gratitude, love, nostalgia, ennui, and so forth, are acquired through socializing agents who define and label such emotions while the individual is experiencing the autonomic reactions of one of the "primaries." Hence, it is argued here, guilt is a socialized response to arousal of the physiological conditions of fear; shame to those of anger; pride to those of satisfaction; and so on.

Paper: Neuropsychology: How Many Emotions Are There?
Reference: Dubois and Adolphs (2015)
Excerpt: Psychological theories disagree on how we attribute emotions to people. A new neuroimaging study shows that such attributions involve a large number of abstract features, rather than a small set of emotion categories.

Paper: Evidence for a three-factor theory of emotions
Reference: Russel and Mehrabian (1977)
Excerpt: Abstract Two studies provided evidence that three independent and bipolar dimensions, pleasure-displeasure, degree of arousal, and dominance-submissiveness, are both necessary and sufficient to adequately define emotional states. In one study with 200 subjects, 42 verbal-report emotion scales were explored in regression analyses as functions of the three dimensions plus a measure of acquiescence bias. Multiple correlation coefficients showed that almost all of the reliable variance in the 42 scales had been accounted for. The specific definitions provided by these equations were replicated in a second study that employed 300 subjects' ratings of 151 emotion-denoting terms on semantic differential-type scales.

Paper: Patterns of cognitive appraisal in emotion.
Reference: Smith and Ellsworth (1985)
Excerpt: There has long been interest in describing emotional experience in terms of underlying dimensions, but traditionally only two dimensions, pleasantness and arousal, have been reliably found. The reasons for these findings are reviewed, and integrating this review with two recent theories of emotions (Roseman, 1984; Scherer, 1982), we propose eight cognitive appraisal dimensions to differentiate emotional experience. In an investigation of this model, subjects recalled past experiences associated with each of 15 emotions, and rated them along the proposed dimensions. Six orthogonal dimensions, pleasantness, anticipated effort, certainty, attentional activity, self-other responsibility/control, and situational control, were recovered, and the emotions varied systematically along each of these dimensions, indicating a strong relation between the appraisal of one's circumstances and one's emotional state.

Ideal answer to the question "How many emotions are there?" based on the papers above using references: One paper claims that there are four emotions (Kemper, 1987), while the others claim that there are three dimensions of emotions (Russel & Mehrabian, 1977), six dimensions of emotions (Smith & Ellsworth, 1985), or that the notion of emotion categories doesn't make sense at all (Dubois & Adolphs, 2015).""",
]

PROMPT_FORMAT = """Question: "{question}"

Here are the relevant papers, and excerpts from those papers:

{papers_str}

Ideal answer to the question "{question}" based on the papers above using references:"""

PAPER_FORMAT = """Paper: {title}
Reference: {reference}
Excerpt: {abstract}"""


def _get_reference(authors: list[str], year: int | None) -> str:
    if len(authors) == 0:
        return f"({year})"
    if len(authors) == 1:
        return f"{authors[0]} ({year})"
    elif len(authors) == 2:
        return f"{authors[0]} and {authors[1]} ({year})"
    else:
        return f"{authors[0]} et al. ({year})"


async def synthesize(question: str, abstracts: list[Abstract]) -> str:

    papers_str = "\n\n".join(
        [
            PAPER_FORMAT.format(
                title=abstract.title,
                reference=_get_reference(abstract.authors, abstract.year),
                abstract=abstract.text,
            )
            for abstract in abstracts
        ]
    )

    suffix = PROMPT_FORMAT.format(
        question=question,
        papers_str=papers_str,
    )

    sep = "\n\n###\n\n"

    prompt = PREFIX + sep + sep.join(SHOTS) + sep + suffix
    number_of_shots = len(SHOTS)
    while (
        num_tokens(prompt) > MAX_TOKENS - MIN_COMPLETION_TOKENS and number_of_shots > 0
    ):
        number_of_shots -= 1
        prompt = PREFIX + sep + sep.join(SHOTS[:number_of_shots]) + sep + suffix

    remaining_tokens = MAX_TOKENS - num_tokens(prompt)

    completion = await recipe.agent().complete(
        prompt=prompt, max_tokens=remaining_tokens, stop="<|endoftext|>"
    )

    return completion


async def synthesize_cli():
    question = "what is the relationship between income and smoking?"
    abstracts = DEFAULT_ABSTRACTS
    gold_standard = "All four papers found that lower income is associated with higher smoking prevalence (Brunilda Casetta et al., 2016; M Huisman et al., 2005; Ariel Esteban Bardach et al., 2016; Mall Leinsalu et al., 2011). Two papers found that less education is also associated with higher smoking prevalence (M Huisman et al., 2005; Mall Leinsalu et al., 2011)."
    print(f"\nGold standard: {gold_standard}\n")
    return await synthesize(question, abstracts)


recipe.main(synthesize_cli)



from collections.abc import Sequence
from dataclasses import dataclass
from statistics import mean

import pandas as pd

from pydantic import BaseModel
from rich import box
from rich.table import Table

from ice.evaluation.evaluate_recipe_result import EvaluatedRecipeResult
from ice.evaluation.utils import CSVS_PATH
from ice.evaluation.utils import precision_score
from ice.evaluation.utils import recall_score
from ice.evaluation.utils import start_time
from ice.utils import latest_commit_hash

SUBSTANTIVELY_CORRECT_MIN_ANSWER_RATING = 4


def as_percent(decimal: float | None) -> str | None:
    if decimal is None:
        return None
    return "{:.0%}".format(decimal)


@dataclass
class ClassificationSummary:
    num_correct: int
    num_evaluated: int
    classification_options: set[str]

    @property
    def proportion_correct(self) -> float | None:
        if self.num_evaluated == 0:
            return None

        return self.num_correct / self.num_evaluated

    def labels_str(self) -> str:
        return ", ".join(list(self.classification_options))


@dataclass
class AnsweredWhenShouldHaveSummary:
    gs_should_answer: list[bool]
    answered: list[bool]

    @property
    def precision(self) -> float:
        if not self.gs_should_answer:
            return 0.0
        return precision_score(self.gs_should_answer, self.answered)

    @property
    def recall(self) -> float:
        if not self.gs_should_answer:
            return 0.0
        return recall_score(self.gs_should_answer, self.answered)

    @property
    def proportion_of_gs_where_should_answer(self) -> float:
        return len(
            [should_answer for should_answer in self.gs_should_answer if should_answer]
        ) / (len(self.gs_should_answer) or 1)


class EvaluationReport(BaseModel):
    technique_name: str
    results: Sequence[EvaluatedRecipeResult]

    @property
    def mean_proportion_gs_found_across_papers(self) -> float | None:
        paper_scores = [
            result.evaluated_excerpts.proportion_gold_standards_found
            for result in self.results
            if result.evaluated_excerpts.proportion_gold_standards_found is not None
        ]

        if len(paper_scores) == 0:
            return None
        return float(mean(paper_scores))

    @property
    def proportion_papers_with_gs_found(self) -> float | None:
        gs_found = [
            result.evaluated_excerpts.num_gold_standards_found > 0
            for result in self.results
            if result.evaluated_excerpts.num_gold_standards_found is not None
        ]

        if len(gs_found) == 0:
            return None

        return mean(gs_found)

    @property
    def classification_summaries(self) -> list[ClassificationSummary]:
        classification_summaries = []

        for i in range(
            0,
            max([len(result.evaluated_classifications) for result in self.results]),
        ):
            is_corrects = []

            for result in self.results:
                try:
                    is_corrects.append(result.evaluated_classifications[i].is_correct)
                except IndexError:
                    continue

            evaluations = [
                is_correct for is_correct in is_corrects if is_correct is not None
            ]

            classification_summaries.append(
                ClassificationSummary(
                    num_correct=sum(evaluations),
                    num_evaluated=len(evaluations),
                    classification_options=self.classification_options(i),
                )
            )

        return classification_summaries

    @property
    def answered_when_should_have(self) -> AnsweredWhenShouldHaveSummary:
        results_to_eval = [
            result for result in self.results if result.gold_standard is not None
        ]
        gs_should_answer = [
            False if result.gold_standard.answer == "" else True  # type: ignore
            for result in results_to_eval
        ]
        answered = [
            False if result.answer == "" else True for result in results_to_eval
        ]

        return AnsweredWhenShouldHaveSummary(
            gs_should_answer=gs_should_answer,
            answered=answered,
        )

    @property
    def elicit_commit(self):
        elicit_commits = list(set([result.elicit_commit for result in self.results]))
        if len(elicit_commits) > 1:
            raise ValueError("Attempting to evaluate multiple Elicit commits at once")
        return elicit_commits[0]

    @property
    def answer_ratings(self):
        return [
            result.answer_rating
            for result in self.results
            if result.answer_rating is not None
        ]

    @property
    def answers_rated(self):
        return len(self.answer_ratings)

    @property
    def average_answer_rating(self):
        return mean(self.answer_ratings) if len(self.answer_ratings) > 0 else None

    @property
    def proportion_of_answers_rated_correct(self):
        return (
            (
                len(
                    [
                        answer_rating
                        for answer_rating in self.answer_ratings
                        if answer_rating >= SUBSTANTIVELY_CORRECT_MIN_ANSWER_RATING
                    ]
                )
                / len(self.answer_ratings)
            )
            if self.answer_ratings
            else None
        )

    @property
    def failure_modes(self):
        # TODO: use more_itertools.flatten instead
        all_failure_modes = []
        for result in self.results:
            if result.failure_modes is None:
                continue
            for failure_mode in result.failure_modes:
                all_failure_modes.append(failure_mode)

        if not all_failure_modes:
            return None

        return pd.Series(all_failure_modes).value_counts().to_string()

    def classification_options(self, classification_idx=int) -> set[str]:
        classification_options: set[str] = set()

        for result in self.results:
            try:
                gold_classification = result.evaluated_classifications[
                    classification_idx
                ].gold
                if gold_classification is not None:
                    classification_options.add(gold_classification)
            except IndexError:
                continue

        return classification_options

    def questions_str(self) -> str:
        questions = list(set([result.question_short_name for result in self.results]))
        return ", ".join(questions)

    def classification_summary_stats_str(self) -> str:
        return "\n".join(
            [
                (
                    f"- Classification {i+1}:"
                    + "\n    - proportion correct: "
                    + (
                        f"{classification_summary.proportion_correct:.0%} ({classification_summary.num_correct} / {classification_summary.num_evaluated})"
                        if classification_summary.proportion_correct is not None
                        else "Nothing to evaluate"
                    )
                    + f"\n    - labels: {classification_summary.labels_str()}"
                )
                for i, classification_summary in enumerate(
                    self.classification_summaries
                )
            ]
        )

    def make_comparison_table(
        self,
        document_id_rows: list[str],
        gold_standard_rows: list[str],
        model_rows: list[str],
    ) -> Table:
        table = Table(width=120, box=box.MARKDOWN)
        table.add_column("Document", no_wrap=True)
        table.add_column("Gold Standard")
        table.add_column("Model")

        for did, gs, mr in zip(document_id_rows, gold_standard_rows, model_rows):
            table.add_row(str(did), str(gs), str(mr))

        return table

    def description_results_table(self) -> Table:
        document_ids = []
        gold_standard_descriptions = []
        model_descriptions = []
        for result in self.results:
            gold_standard = result.gold_standard
            if gold_standard is not None:
                if gold_standard.answer != gold_standard.classifications[0]:
                    document_ids.append(gold_standard.document_id)
                    gold_standard_descriptions.append(gold_standard.answer)
                    model_descriptions.append(result.answer)
        return self.make_comparison_table(
            document_ids, gold_standard_descriptions, model_descriptions
        )

    def classification_results_table(self) -> Table:
        document_ids = []
        gold_standard_classifications = []
        model_classifications = []
        for result in self.results:
            gold_standard = result.gold_standard
            if gold_standard is not None:
                document_ids.append(gold_standard.document_id)
                gold_standard_classifications.append(
                    gold_standard.classifications[1] or ""
                )
                model_classification = (
                    result.classifications[0] if result.classifications else ""
                )
                model_classifications.append(model_classification or "")
        return self.make_comparison_table(
            document_ids, gold_standard_classifications, model_classifications
        )

    def excerpts_summary_stats_str(self) -> str:
        s = ""
        if self.mean_proportion_gs_found_across_papers:
            s += f"### Average % of gold standards found: \n\n{self.mean_proportion_gs_found_across_papers:.0%}"
        if self.proportion_papers_with_gs_found:
            s += f"\n\n### Proportion of papers with at least one gold standard found: \n\n{self.proportion_papers_with_gs_found:.0%}"
        if not s:
            s = "n/a"
        return s

    def precision_recall_str(self) -> str:
        return f"""- precision: {self.answered_when_should_have.precision:.0%}
- recall: {self.answered_when_should_have.recall:.0%}
- (proportion of gold standards where should answer: {self.answered_when_should_have.proportion_of_gs_where_should_answer:.0%})
"""

    def to_rich_elements(self) -> list[str | Table]:
        return [
            f"""
# Performance on papers

{"---".join(str(r) for r in self.results)}

# Summary of results

## Classifications:
""",
            self.classification_results_table(),
            """
## Descriptions:
""",
            self.description_results_table(),
            """
## Classification summary stats:
""",
            self.classification_summary_stats_str(),
            """
## Excerpts summary stats:
""",
            self.excerpts_summary_stats_str(),
            """
## Answered when should have:
""",
            self.precision_recall_str(),
        ]

    def __str__(self) -> str:
        return "\n".join(str(e) for e in self.to_rich_elements())

    def make_dashboard_row_df(self):
        CSVS_PATH.mkdir(parents=True, exist_ok=True)

        questions = self.questions_str()
        row = {
            "Questions": questions,
            "Technique": self.technique_name,
            "Eval date": start_time,
            "Elicit commit": self.elicit_commit,
            "ICE commit": latest_commit_hash(),
            "Splits": ", ".join(
                set(
                    result.gold_standard.split
                    for result in self.results
                    if result.gold_standard and result.gold_standard.split
                )
            ),
            "# of results": len(self.results),
            "Papers with >= 1 GS quote found": as_percent(
                self.proportion_papers_with_gs_found
            ),
            "Mean % of GS quotes found": as_percent(
                self.mean_proportion_gs_found_across_papers
            ),
            "Precision (answer if should)": as_percent(
                self.answered_when_should_have.precision
            ),
            "Recall (answer if should)": as_percent(
                self.answered_when_should_have.recall
            ),
            "% GS where should answer": as_percent(
                self.answered_when_should_have.proportion_of_gs_where_should_answer
            ),
            "Answers rated": self.answers_rated,
            "Average answer rating": self.average_answer_rating,
            "% answers rated substantively correct": as_percent(
                self.proportion_of_answers_rated_correct
            ),
            "Failure mode counts": self.failure_modes,
        }
        for i, classification_summary in enumerate(self.classification_summaries):
            row[f"Classification {i+1} labels"] = classification_summary.labels_str()

            row[f"Classification {i+1} correct"] = as_percent(
                classification_summary.proportion_correct
            )

            row[
                f"Classification {i+1} # evaluated"
            ] = classification_summary.num_evaluated

        df = pd.DataFrame([row])
        df.to_csv(
            CSVS_PATH / f"dashboard_row {start_time} {questions}.csv",
            index=False,
        )
        return df

    def make_experiments_evaluation_df(self):
        CSVS_PATH.mkdir(parents=True, exist_ok=True)

        rows = []
        questions = self.questions_str()

        for result in self.results:
            row = {
                "question_short_name": result.question_short_name,
                "technique": self.technique_name,
                "eval_date": start_time,
                "elicit_commit": result.elicit_commit,
                "ice_commit": latest_commit_hash(),
                "document_id": result.document_id,
                "split": result.gold_standard.split if result.gold_standard else None,
                "experiment": result.gold_standard.experiment
                if result.gold_standard
                else None,
                "total_gs_quotes": len(
                    result.evaluated_excerpts.gold_standards_in_excerpts_results
                ),
                "gs_quotes_found": result.evaluated_excerpts.num_gold_standards_found,
                "proportion_gs_quotes_found": result.evaluated_excerpts.proportion_gold_standards_found,
                "evaluated_excerpts": str(result.evaluated_excerpts),
                "excerpts": result.evaluated_excerpts.excerpts,
                "gs_quotes": result.evaluated_excerpts.gold_standards_str(),
                "answer": result.answer,
                "gs_answer": result.gold_standard.answer
                if result.gold_standard
                else None,
                "answer_rating": result.answer_rating,
                "failure_modes": result.failure_modes,
            }

            if result.gold_standard:
                row["gs_answer"] = result.gold_standard.answer
                row["should_answer"] = result.gold_standard.answer != ""
                row["answered"] = result.answer != ""

            for i, classification in enumerate(result.evaluated_classifications):
                row[f"classification_{i+1}_correct"] = classification.is_correct
                row[f"classification_{i+1}"] = classification.predicted
                row[f"gold_classification_{i+1}"] = classification.gold

            rows.append(row)
        df = pd.DataFrame(rows)
        df.to_csv(
            CSVS_PATH / f"experiments_evaluation {start_time} {questions}.csv",
            index=False,
        )
        return df



import httpx

from fvalues import F

from ice.recipe import recipe


async def search(query: str) -> dict:
    async with httpx.AsyncClient() as client:
        params = {"q": query, "hl": "en", "gl": "us", "api_key": "e29...b4c"}
        response = await client.get("https://serpapi.com/search", params=params)
        return response.json()


def render_results(data: dict) -> str:
    if not data or not data.get("organic_results"):
        return "No results found"

    results = []
    for result in data["organic_results"]:
        title = result.get("title")
        link = result.get("link")
        snippet = result.get("snippet")
        if not title or not link or not snippet:
            continue
        results.append(F(f"{title}\n{link}\n{snippet}\n"))

    return F("\n").join(results)


async def search_string(
    question: str = "Who is the president of the United States?",
) -> str:
    results = await search(question)
    return render_results(results)


recipe.main(search_string)



import importlib.util
import os.path

from typing import Type

from ice.environment import env
from ice.recipe import Recipe
from ice.recipes import get_recipe_classes


def is_filename(name: str) -> bool:
    """
    Check if name is a valid filename with a .py extension
    """
    return os.path.isfile(name) and name.endswith(".py")


def load_recipe_class_from_file(filename: str) -> Type[Recipe]:
    # Load the recipe class from the specified file using importlib
    spec = importlib.util.spec_from_file_location("recipe_module", filename)
    if spec is None:
        raise ValueError(f"Could not load recipe from file {filename}")
    module = importlib.util.module_from_spec(spec)
    loader = spec.loader
    if loader is None:
        raise ValueError(f"Could not load recipe from file {filename}")
    loader.exec_module(module)
    # Find the last subclass of Recipe in the module
    recipe_class = None
    for name, obj in vars(module).items():
        try:
            if issubclass(obj, Recipe) and obj is not Recipe:
                recipe_class = obj
        except TypeError:
            # obj is not a class, skip it
            continue
    # Raise an exception if no recipe class is found
    if recipe_class is None:
        raise ValueError(f"No recipe class found in {filename}")
    return recipe_class


def load_recipe_class_by_name(recipe_name: str) -> Type[Recipe]:
    recipe_classes = get_recipe_classes()
    try:
        recipe_class = next(
            r
            for r in recipe_classes
            if r.__name__.lower().startswith(recipe_name.lower())
        )
    except StopIteration:
        raise ValueError(
            f"Recipe '{recipe_name}' not found in {[r.__name__ for r in recipe_classes]}"
        )
    return recipe_class


async def ask_user_for_recipe_class() -> Type[Recipe]:
    recipe_classes = get_recipe_classes()
    recipe_names = [r.__name__ for r in recipe_classes]
    recipe_name = await env().select("Recipe", recipe_names)
    recipe_class = [r for r in recipe_classes if r.__name__ == recipe_name][0]
    return recipe_class


async def select_recipe_class(*, recipe_name: str | None = None) -> Type[Recipe]:
    if recipe_name is not None:
        if is_filename(recipe_name):
            recipe_class = load_recipe_class_from_file(recipe_name)
        else:
            recipe_class = load_recipe_class_by_name(recipe_name)
    else:
        recipe_class = await ask_user_for_recipe_class()
    return recipe_class



from collections.abc import Callable
from collections.abc import Sequence
from functools import partial

from ice.apis.openai import TooLongRequestError
from ice.formatter.transform.value import numbered_list
from ice.paper import Paper
from ice.recipe import recipe
from ice.recipes.meta.eval_paper_qa.quick_list import quick_list
from ice.recipes.meta.eval_paper_qa.types import PaperQaAnswer
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard
from ice.recipes.meta.eval_paper_qa.types import PaperQaMethod
from ice.recipes.meta.eval_paper_qa.utils import convert_demonstration_example
from ice.recipes.meta.eval_paper_qa.utils import identify_gs_str
from ice.recipes.primer.paper_qa import answer_for_paper
from ice.recipes.primer.qa import answer
from ice.recipes.program_search.nodes.answer.answer import demonstration_answer
from ice.recipes.program_search.nodes.answer.answer import (
    demonstration_answer_with_reasoning,
)
from ice.recipes.program_search.nodes.answer.answer import elicit_answer_prompt
from ice.recipes.program_search.nodes.answer.types import Demonstration
from ice.recipes.program_search.nodes.select.select import as_bool
from ice.recipes.program_search.nodes.select.select import (
    select_results_using_top_monot5_paragraph,
)
from ice.utils import map_async


async def _cheating_qa_baseline(
    paper: Paper,
    question: str,
    gold_support: Sequence[str] | None,
    enumerate_answer: bool,
):
    """Baseline that uses gold standard support, without searching the paper.
    (hence "cheating")

    Args:
        paper (Paper): _description_
        question (str): _description_
        gold_support (Sequence[str] | None): _description_
        enumerate_answer (bool, optional): _description_. Defaults to False.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    relevant_str = "\n\n".join(gs for gs in gold_support) if gold_support else ""
    if not relevant_str:
        raise ValueError("Method requires gold support")
    response: str | Sequence[str] = await answer(
        context=relevant_str, question=question
    )
    if enumerate_answer:
        response = await quick_list(question, str(response))
    assert gold_support
    return PaperQaAnswer(
        answer=response,
        support_candidates=gold_support,
        support_labels=[True for _ in gold_support],
    )


cheating_qa_baseline_str_answer: PaperQaMethod[str] = partial(
    _cheating_qa_baseline, enumerate_answer=False
)
cheating_qa_baseline_list_answer: PaperQaMethod[Sequence[str]] = partial(
    _cheating_qa_baseline, enumerate_answer=True
)


async def _paper_qa_baseline(
    paper: Paper, question, gold_support: Sequence[str] | None, enumerate_answer: bool
) -> PaperQaAnswer:
    gold_support  # unused
    answer: str | Sequence[str]
    all_paras = [str(p).casefold().strip() for p in paper.paragraphs if str(p).strip()]
    answer, predicted_paras = await answer_for_paper(paper, question, top_n=1)
    predicted_paras = [p.casefold().strip() for p in predicted_paras]
    if enumerate_answer:
        answer = await quick_list(question, answer)
    return PaperQaAnswer(
        answer=answer,
        support_candidates=all_paras,
        support_labels=[p.casefold().strip() in predicted_paras for p in all_paras],
    )


async def _elicit_paper_qa_baseline(
    paper: Paper, question, gold_support: Sequence[str] | None
) -> PaperQaAnswer:
    gold_support  # unused
    answer: str | Sequence[str]
    all_paras = [str(p).casefold().strip() for p in paper.paragraphs if str(p).strip()]
    top_paragraph = await select_results_using_top_monot5_paragraph(
        question=question,
        texts=all_paras,
    )
    answer = await elicit_answer_prompt(question, top_paragraph)
    return PaperQaAnswer(
        answer=answer,
        support_candidates=all_paras,
        support_labels=[(p.casefold().strip() == top_paragraph) for p in all_paras],
    )


paper_qa_baseline_str_answer: PaperQaMethod[str] = partial(
    _paper_qa_baseline, enumerate_answer=False
)
paper_qa_baseline_list_answer: PaperQaMethod[Sequence[str]] = partial(
    _paper_qa_baseline, enumerate_answer=True
)


async def _demonstration_answer(
    question: str,
    texts: Sequence[str],
    demonstrations: Sequence[Demonstration],
    reasoning: bool,
):
    try:
        if not reasoning:
            return await demonstration_answer(
                question=question, texts=texts, demonstrations=demonstrations
            )
        else:
            return await demonstration_answer_with_reasoning(
                question=question, texts=texts, demonstrations=demonstrations
            )
    except TooLongRequestError:
        demonstrations = demonstrations[:-1]
        if len(demonstrations) < 1:
            raise
        return await _demonstration_answer(
            question=question,
            texts=texts,
            demonstrations=demonstrations,
            reasoning=reasoning,
        )


async def preselected_few_shot_qa_baseline(
    paper: Paper,
    question: str,
    gold_support: Sequence[str] | None,
    enumerate_answer: bool,
    few_shot_demonstration_func: Callable[[str], Sequence[PaperQaGoldStandard]],
    selections: Sequence[str] | None = None,
    paper_division_func: Callable[[Paper], Sequence[str]] | None = None,
    reasoning: bool = False,
):
    demonstration_examples = few_shot_demonstration_func(paper.document_id)
    if paper_division_func:
        demonstration_examples = await map_async(
            demonstration_examples,
            partial(
                convert_demonstration_example,
                paper_division_func=paper_division_func,
            ),
        )
    demonstration_examples = [ex for ex in demonstration_examples if ex.gold_support][
        :3
    ]
    demonstrations = [
        Demonstration(
            question=g.question,
            texts=g.gold_support,
            answer=g.gold_answer
            if isinstance(g.gold_answer, str)
            else numbered_list(g.gold_answer).transform(),
        )
        for g in demonstration_examples
    ]
    if paper_division_func and gold_support:
        gold_support = await identify_gs_str(paper_division_func(paper), gold_support)
    support = gold_support or selections
    assert support
    answer = await _demonstration_answer(
        question=question,
        texts=support,
        demonstrations=demonstrations,
        reasoning=reasoning,
    )

    if enumerate_answer:
        answer = await quick_list(question, answer)
    if paper_division_func:
        support_candidates = paper_division_func(paper)
        support_labels = as_bool(support, support_candidates)
    else:
        support_labels = [False for _ in support]

    return PaperQaAnswer(
        answer=answer,
        support_candidates=support,
        support_labels=support_labels,
    )


# Refactoring notes
# for evaling more structured recipe return types, pass in the evaluation function
# specify the question short name; we need somewhere a mapping of these to the ModelType
#

# to specify an experiment from the command line (or perhaps better a file)
# - recipe function (paper, question, gold_support | None) -> PaperQaAnswer (answer, support)
# - generator of (paper, question, gold_support | None) from parsed golds (question_and_answer_func)
# - split (+ slice?) and/or files
# - question_short_name (has an implied ModelType that must match question_and_answer_func)
# - eval function (where the return type matches the output of the recipe function)


recipe.main(cheating_qa_baseline_str_answer)



from structlog import get_logger

from ice.recipe import recipe
from ice.recipes.elicit.common import send_elicit_request

log = get_logger()


def make_request_body(
    query: str, num_papers: int = 4, filters: dict | None = None
) -> dict:
    """
    Make the request body for the Elicit search endpoint.
    """
    if filters is None:
        filters = {}
    return dict(
        query=query,
        start=0,
        stop=num_papers,
        qaColumns=[],
        filters=filters,
    )


async def elicit_search(
    question: str = "What is the effect of creatine on cognition?",
    num_papers: int = 4,
    # TODO: Dynamically consult https://elicit.org/api/backend
    endpoint: str = "https://prod.elicit.org/elicit-red/lit-review",
):
    """
    Search Elicit for papers related to a question.
    """
    log.info(f"Searching Elicit for query: {question}, endpoint: {endpoint}")

    filters = None

    request_body = make_request_body(
        query=question, num_papers=num_papers, filters=filters
    )

    response = send_elicit_request(
        request_body=request_body,
        endpoint=endpoint,
    )
    return response


recipe.main(elicit_search)



import itertools
import os
import subprocess
import threading as td
import time

from collections import defaultdict
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Coroutine
from collections.abc import Iterable
from collections.abc import Sequence
from enum import Enum
from functools import cache
from random import Random
from typing import Any
from typing import Generic
from typing import TypeVar

import anyio
import tqdm

from more_itertools import windowed
from structlog.stdlib import get_logger
from transformers import GPT2TokenizerFast

log = get_logger()


def _merge(recurse, path: list, base: dict, nxt: dict) -> dict:
    for k, v in nxt.items():
        if k not in base:
            base[k] = v
        else:
            base[k] = recurse(recurse, path + [k], base[k], v)
    return base


def deep_merge(base, nxt):
    """
    Performs a *limited* deep merge of nxt into base.
    Type differences are overriden by nxt.
    Lists are extended, but elements are not changed or recursed into.
    Sets are unioned but not recursed into.
    """

    def merge_strategy(
        merge_strategy,
        path: list,
        base,
        nxt,
    ):
        if not (isinstance(base, type(nxt)) or isinstance(nxt, type(base))):
            return nxt
        elif isinstance(nxt, dict):
            return _merge(merge_strategy, path, base, nxt)
        elif isinstance(nxt, list) or isinstance(nxt, tuple):
            return base + nxt
        elif isinstance(nxt, set):
            return base | nxt
        return nxt

    return merge_strategy(merge_strategy, [], base, nxt)


InputType_co = TypeVar("InputType_co", covariant=True)
ReturnType_co = TypeVar("ReturnType_co", covariant=True)


# inspired by http://bluebirdjs.com/docs/api/promise.map.html
async def map_async(
    input_list: Sequence[InputType_co],
    fn: Callable[[InputType_co], Coroutine[Any, Any, ReturnType_co]],
    max_concurrency: int | None = None,
    semaphore: anyio.Semaphore | None = None,
    show_progress_bar: bool = False,
) -> list[ReturnType_co]:
    result_boxes: list[list[ReturnType_co]] = [[] for _ in input_list]

    if not semaphore:
        semaphore = anyio.Semaphore(max_concurrency or len(input_list))

    if show_progress_bar:
        progress_bar = tqdm.tqdm(total=len(input_list))

    async def box_result(
        input: Any, result_box: list[ReturnType_co], semaphore: anyio.Semaphore
    ) -> None:
        async with semaphore:
            result = await fn(input)
        result_box.extend([result])

        if show_progress_bar:
            progress_bar.update(1)

    async with anyio.create_task_group() as tg:
        for i in range(len(input_list)):
            tg.start_soon(box_result, input_list[i], result_boxes[i], semaphore)

    if show_progress_bar:
        progress_bar.close()

    return [result_box[0] for result_box in result_boxes]


async def filter_async(
    iterable: Iterable[InputType_co],
    fn: Callable[[InputType_co], Coroutine[Any, Any, bool]],
    max_concurrency: int | None = None,
    show_progress_bar: bool = False,
) -> list[InputType_co]:
    iterable_list = list(iterable)
    values = await map_async(
        iterable_list,
        fn,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    return [item for item, value in zip(iterable_list, values) if value]


T = TypeVar("T")
S = TypeVar("S")


async def reduce_async(
    fn: Callable[[T, S], Awaitable[T]], iterable: Iterable[S], initial: T
):
    current = initial
    for item in iterable:
        current = await fn(current, item)
    return current


class _Sentinel(Enum):
    token = 0


_sentinel = _Sentinel.token


def window_dropping(items: Sequence[T], n, step) -> Sequence[Sequence[T]]:
    """Windows over items, shortening n if necessary"""
    return [
        [i for i in window if i is not _sentinel]
        for window in windowed(items, n=n, step=step, fillvalue=_sentinel)
    ]


def longest_common_prefix(xs: Sequence[str]) -> str:
    if not xs:
        return ""
    prefix = xs[0]
    for x in xs[1:]:
        prefix = os.path.commonprefix([prefix, x])
    return prefix


AsyncComparator = Callable[[T, T], Awaitable[int]]


async def _nsmallest_async(
    n: int,
    items: list[T],
    cmp: AsyncComparator,
    semaphore: anyio.Semaphore | None,
    offset: int = 0,
    seed: int = 1,
) -> list[T]:
    n = max(0, n)
    if len(items) <= 1 or n == 0:
        return items[:n]

    pivot = items.pop(Random(seed).randrange(len(items)))

    async def get_key(item: T) -> bool:
        return await cmp(item, pivot) >= 0

    # partitions[False] contains items < pivot, partitions[True] contains items >= pivot
    partitions = defaultdict(list)
    keys = await map_async(items, get_key, semaphore=semaphore)
    for item, key in zip(items, keys):
        partitions[key].append(item)

    async def recurse(key: bool) -> tuple[bool, list[T]]:
        delta = len(partitions[False]) + 1 if key else 0
        # Represent the path to this node in the call tree as a binary number, and use
        # that as the seed.
        new_seed = 2 * seed + int(key)
        return key, await _nsmallest_async(
            n - delta, partitions[key], cmp, semaphore, offset + delta, new_seed
        )

    partitions = defaultdict(list, await map_async(list(partitions), recurse))
    return (partitions[False] + [pivot] + partitions[True])[:n]


async def nsmallest_async(
    n: int,
    items: Iterable[T],
    cmp: AsyncComparator,
    max_concurrency: int | None = None,
) -> list[T]:
    semaphore = anyio.Semaphore(max_concurrency) if max_concurrency else None
    return await _nsmallest_async(n, list(items), cmp, semaphore)


def flatten(xs: Iterable[Iterable[T]]) -> list[T]:
    return list(itertools.chain(*xs))


def chunk_by(xs: list[T], n: int, f: Callable[[T], float]) -> list[list[T]]:
    """
    Split a list into chunks of size n, with the size of each element of l computed by f.
    """
    chunks = []
    current_chunk: list[T] = []
    current_chunk_size: float = 0.0
    for x in xs:
        x_size = f(x)
        if current_chunk_size + x_size > n:
            chunks.append(current_chunk)
            current_chunk = []
            current_chunk_size = 0.0
        current_chunk.append(x)
        current_chunk_size += x_size
    if current_chunk:
        chunks.append(current_chunk)
    return chunks


def quoted(multiline_string: str) -> str:
    return "\n".join(f"> {line}" for line in multiline_string.split("\n"))


ArgT = TypeVar("ArgT")
ReturnT_co = TypeVar("ReturnT_co", covariant=True)


class DynamicBatcher(Generic[ArgT, ReturnT_co]):
    # Shamelessly adapted/simplified/typed based on:
    # https://github.com/cortexlabs/nucleus/blob/master/src/cortex/cortex_internal/lib/api/utils.py
    # Apache 2.0 license
    # Copyright 2022 Cortex Labs, Inc.
    # Thanks, Cortex!
    def __init__(
        self,
        handler: Callable[[list[ArgT]], list[ReturnT_co]],
        max_batch_size: int,
        batch_interval_seconds: int,
        test_mode: bool = False,
    ):
        self.handler = handler

        self.batch_max_size = max_batch_size
        self.batch_interval_seconds = batch_interval_seconds  # measured in seconds
        self.test_mode = test_mode  # only for unit testing
        self._test_batch_lengths: list[int] = []  # only when unit testing

        self.barrier = td.Barrier(self.batch_max_size + 1)

        self.samples: dict[int, ArgT] = {}
        self.results: dict[int, Exception | ReturnT_co] = {}
        td.Thread(target=self._batch_engine, daemon=True).start()

        self.sample_id_generator = itertools.count()

    def _batch_engine(self):
        while True:
            if len(self.results) > 0:
                time.sleep(0.001)
                continue

            try:
                self.barrier.wait(self.batch_interval_seconds)
            except td.BrokenBarrierError:
                pass

            self.results = {}
            sample_ids = self._get_sample_ids(self.batch_max_size)
            try:
                if self.samples:
                    batch = self._make_batch(sample_ids)

                    results = self.handler(batch)
                    if not isinstance(results, list):
                        raise RuntimeError(
                            f"please return a list when using server side batching, got {type(results)}"
                        )

                    if self.test_mode:
                        self._test_batch_lengths.append(len(results))

                    self.results = dict(zip(sample_ids, results))
            except Exception as e:
                self.results = {sample_id: e for sample_id in sample_ids}
                log.exception("Unhandled exception in server-side batching")
            finally:
                for sample_id in sample_ids:
                    del self.samples[sample_id]
                self.barrier.reset()

    def _get_sample_ids(self, max_number: int) -> list[int]:
        if len(self.samples) <= max_number:
            return list(self.samples.keys())
        return sorted(self.samples)[:max_number]

    def _make_batch(self, sample_ids: list[int]) -> list[ArgT]:
        batched_samples: list[ArgT] = []
        for sample_id in sample_ids:
            batched_samples.append(self.samples[sample_id])

        return batched_samples

    def _enqueue_request(self, sample_id: int, kwargs: ArgT):
        """
        Enqueue sample for batch processing. This is a blocking method.
        """

        self.samples[sample_id] = kwargs
        try:
            self.barrier.wait()
        except td.BrokenBarrierError:
            pass

    def process(self, arg: ArgT) -> ReturnT_co:
        """
        Queues a request to be batched with other incoming request, waits for the response
        and returns the processed result. This is a blocking method.
        """
        sample_id = next(self.sample_id_generator)
        self._enqueue_request(sample_id, arg)
        result = self._get_result(sample_id)
        return result

    def _get_result(self, sample_id: int) -> ReturnT_co:
        """
        Return the processed result. This is a blocking method.
        """
        while sample_id not in self.results:
            time.sleep(0.001)

        result = self.results[sample_id]
        del self.results[sample_id]

        if isinstance(result, Exception):
            raise result
        else:
            return result


def truncate_by_tokens(text: str, *, max_tokens: int = 7500):
    # Full context is 8000 tokens * ~3.5 chars/token
    chars = len(text)
    max_chars = int(max_tokens * 3.5)
    if chars > max_chars:
        log.warning(
            f"Truncating from {chars} to {max_chars} characters",
        )
        text = text[:max_chars]
    return text


def window_by_tokens(text: str, *, max_tokens: int = 7000):
    chars = len(text)
    max_chars = int(max_tokens * 3.5)
    for i in range(0, chars, max_chars):
        yield text[i : i + max_chars]


K = TypeVar("K")  # key type

V = TypeVar("V")  # value type


def max_by_value(
    d: dict[K, V], *, key: Callable[[V], Any] = lambda x: x
) -> tuple[K, V]:
    return max(d.items(), key=lambda x: key(x[1]))


@cache
def make_gpt2_tokenizer() -> GPT2TokenizerFast:
    return GPT2TokenizerFast.from_pretrained("gpt2")


def n_tokens(text: str) -> int:
    return len(make_gpt2_tokenizer().tokenize(text))


def latest_commit_hash():
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()



from pydantic import BaseModel


def camel_cased(string: str) -> str:
    return "".join(w.capitalize() if i else w for i, w in enumerate(string.split("_")))


class RouteModel(BaseModel):
    class Config:
        alias_generator = camel_cased
        allow_population_by_field_name = True



import json

from functools import reduce
from pathlib import Path

import httpx

from dotenv import dotenv_values
from tenacity import retry
from tenacity import stop_after_attempt
from tenacity import wait_random_exponential

from ice.cache import diskcache
from ice.settings import settings
from ice.utils import deep_merge


script_dir = Path(__file__).parent
root_dir = script_dir.parent.parent.parent
config = dotenv_values(root_dir / ".env")


@diskcache()
# try 5 times, because sometimes preview apps take a while to start responding again after they have an issue
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(2))
def send_elicit_request(*, request_body, endpoint: str):
    if not settings.ELICIT_AUTH_TOKEN:
        raise ValueError(
            "ELICIT_AUTH_TOKEN not found. Please look it up by checking idToken in cookies for Elicit. The token should NOT start with Bearer, it should just be a string of letters and numbers."
        )

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {settings.ELICIT_AUTH_TOKEN}",
    }
    chunks = []

    with httpx.stream(
        "POST",
        endpoint,
        json=request_body,
        headers=headers,
        timeout=40,
    ) as r:
        for chunk in r.iter_text():
            chunks.append(chunk)

    joined = "".join(chunks)
    split_on_newlines = joined.split("\n")
    as_json = [json.loads(line) for line in split_on_newlines if line]
    merged = reduce(deep_merge, as_json)
    if "detail" in merged:
        raise ValueError(merged["detail"])
    return merged



from structlog.stdlib import get_logger

from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import Recipe
from ice.recipes.rank_paragraphs import RankParagraphs

Experiment = str

log = get_logger()

DEFAULT_QUESTION_SHORT = (
    "What were the trial arms (subgroups of participants) of the experiment?"
)

DEFAULT_QUESTION_LONG = "What were the trial arms (subgroups of participants) of the experiment? List one per line."

DEFAULT_ANSWER_PREFIX = "Answer: The trial arms were:\n-"


def make_qa_prompt(
    question_short: str,
    question_long: str,
    answer_prefix: str,
    paragraphs: list[Paragraph],
) -> str:
    paragraphs_str = "\n\n".join(map(str, paragraphs))
    return f"""
Answer the question "{question_short}" based on the following paragraphs.

Paragraphs:

{paragraphs_str}

Question: {question_long}

{answer_prefix}
""".strip()


class ComparisonsQA(Recipe):
    async def run(
        self,
        paper: Paper,
        question_short: str = DEFAULT_QUESTION_SHORT,
        question_long: str = DEFAULT_QUESTION_LONG,
        num_paragraphs: int = 3,
        answer_prefix: str = DEFAULT_ANSWER_PREFIX,
    ):

        rank_paragraphs = RankParagraphs(mode=self.mode)

        top_paragraphs = await rank_paragraphs.run(
            paper=paper, question=question_short, n=num_paragraphs
        )

        qa_prompt = make_qa_prompt(
            question_short=question_short,
            question_long=question_long,
            answer_prefix=answer_prefix,
            paragraphs=top_paragraphs,
        )

        answer = await self.agent().complete(prompt=qa_prompt, max_tokens=500)

        return answer



from collections.abc import Callable
from collections.abc import Sequence
from collections.abc import Sized

from structlog.stdlib import get_logger

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt

log = get_logger()

# TODO: WRITE PROMPTS ##

NAME_ARMS_PREFACE = """When evaluating a Randomized Controlled Trial, we should first identify its methodology; in particular, we should identify how different subgroups of participants were split into different trial arms, in order to receive different treatments or controls. We will look at a few paragraphs from different papers to identify the trial arms for specific experiments. Once we do this task, we will be able to draw a chart identifying, for each experiment, what the trial arms were. After that, we will look to identify how many participants were put into each arm, and how randomization was performed. Some of these excerpts may not be helpful, so we'll only pay attention to the excerpts that provide explicit information about what the trial arms (subgroups of participants) were for a specific experiment."""

NAME_ARMS_EXAMPLE_TEMPLATE = """
{ordinal}, consider the following excerpt{maybe_plural_excerpts} from a paper that conducted {experiment_count_word} experiment{maybe_plural_experiments}:

{paragraphs}

This paper discusses {experiment_count_word} experiment{maybe_plural_experiments}:

{experiments}

Let's focus on identifying the trial arms just for this experiment for now:

{experiment_in_question}

Let's think about what each excerpt tells us, if anything, about the trial arms (subgroups of participants) for that experiment: {reasoning}

In summary, the trial arms for this experiment in particular were:

{answer}
""".strip()


NAME_ARMS_EXAMPLES: list[dict[str, ValueTransform[Sequence[str]] | str | int]] = [
    dict(
        paragraphs=numbered_list(
            [
                """This study was a 6-month, 2-arm cluster randomized controlled trial comparing a pre-school self-regulation program (PRSIST Program) with typical practice (control group). Fifty pre-school centers in metropolitan and regional areas of Australia were recruited to be broadly representative of population proportions for geography (84% metropolitan), socio-economic decile for their catchment area (M = 5.91, SD = 2.24, range = 1-10), and statutory quality assessment rating (i.e., 44% Exceeding, 48% Meeting, 4% Working Toward, 4% unrated against the National Quality Standard). Australia's early childhood education and care (ECEC) sector includes a range of pre-school provision (e.g., preschool for 4-5-year old children in the year before formal schooling, long-day care services from infant to age 5, family day care) that is delivered by not-for-profit, for-profit or state providers. While there is no state or national curriculum for the Australian ECEC sector, all pre-school services are required to follow the Australian Early Years Learning Framework, which outlines expected outcomes of children from birth to age 5. For this study, participating pre-schools: were structurally equivalent in terms of being long-day care services providing care to children aged 2-5 years, up to 5 days/week; were run by community or not-for-profit providers; and had at least one Bachelor-qualified educator (or government waiver).""",
                """Adherence to intervention participation thresholds was evaluated in terms of educators' completion of the online professional development modules and having engaged children in a minimum of three child activities per week. Engagement with optional program components (i.e., use of formative assessment tool, participation in monthly teleconference calls) was also captured. Educators' engagement in the online professional development was captured via log in and tracking functionality of the professional development modules. Of the 25 intervention centers, 20 services (80%) had at least one educator complete the professional development within the first 3 months of the intervention period (20% of the services had more than one educator complete the professional development during this time). Type and frequency of child activities each month was captured through a custom-designed activity sticker calendar, which was returned monthly to the research team. On average, six of the program's self-regulation activities were facilitated with children each week across the intervention period, ranging from none per week to 22 per week. Further, the charts indicated the suggested diversity of activities was met by most centers in most weeks, and certainly over the duration of the program (by centers who engaged with the child activities).""",
            ]
        ),
        reasoning="""Excerpt 1 identifies the two trial arms as (1) a pre-school self-regulation program (PRSIST Program) and (2) typical practice (control group). Ewxcerpt 2 discuss adherence to protocols but does not add any information about trial arms.""",
        answer=numbered_list(
            [
                "a pre-school self-regulation program (PRSIST Program)",
                "control (typical practice)",
            ]
        ),
        experiments=numbered_list(["the PRSIST program"]),
        experiment_in_question="the PRSIST program",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """to the control group (column 2, rows 15 and 16), whereas for efficiency the observed increase is 46.67% (column 8, same rows). This large difference in efficiency can be also seen in Figure 8 , where the bottom of the post-treatment box of the mindfulness group has almost the same value than the top of the same box of the control group, i.e. after the intervention, approximately 75% of the subjects who practiced mindfulness were more efficient than 75% of the subjects who did not.""",
                """After performing the normality and homoscedasticity tests to determine the applicability of either parametric or nonparametric statistical tests, two mixed-model ANOVAs were carried out for each response variable 7 .""",
                """To avoid any differences in the ISEIS lessons taught to the subjects, all of them had the same professors and the same content taught at the same pace.""",
            ]
        ),
        reasoning="""Excerpt 1 discusses two arms, a mindfulness group and a control group. Excerpt 2 does not add any information about the arms. Excerpt 3 does not add any information about arms.""",
        answer=numbered_list(["mindfulness group", "control group"]),
        experiments=numbered_list(
            [
                "Baseline experiment",
                "1st internal replication",
                "2nd internal replication",
            ]
        ),
        experiment_in_question="2nd internal replication",
    ),
]

NAME_ARMS_SHARED: dict[str, OrdinalWord | DependentTransform[int | Sized]] = dict(
    ordinal=OrdinalWord(capitalize=True, finally_case="Finally"),
    maybe_plural_excerpts=plural_transform(
        key="paragraphs", singular_case="", plural_case="s"
    ),
    experiment_count_word=CountWord("experiments"),
    maybe_plural_experiments=plural_transform(
        key="experiments", singular_case="", plural_case="s"
    ),
)


NAME_ARMS_REASONING_STOP = ("\n\nIn summary",)


def make_name_arms_from_exps(
    experiments: Sequence[str],
    experiment_in_question: str,
) -> Callable[[int], MultipartReasoningPrompt]:
    def make_name_experiments_prompt_func(num_shots: int) -> MultipartReasoningPrompt:
        def name_experiments_prompt(
            paragraphs: Sequence[str],
            helpfulness: str | None = None,
            reasoning: str | None = None,
        ) -> str:
            helpfulness  # ignored
            last_example: dict[
                str, ValueTransform[Sequence[str]] | str | StopSentinel | int
            ] = dict(
                paragraphs=numbered_list(paragraphs),
                reasoning=reasoning if reasoning else stop("Excerpt 1"),
                experiments=numbered_list(experiments),
                experiment_in_question=experiment_in_question,
            )
            if reasoning:
                last_example["answer"] = stop("")
            shots = format_multi(
                NAME_ARMS_EXAMPLE_TEMPLATE,
                NAME_ARMS_EXAMPLES[:num_shots] + [last_example],
                NAME_ARMS_SHARED,
            )
            return "\n\n".join([NAME_ARMS_PREFACE] + list(shots))

        return name_experiments_prompt

    return make_name_experiments_prompt_func


def get_name_arms_reasoning(response: str) -> str:
    return "".join(("Excerpt 1", get_part(response, ": ", "\nIn summary")))



from itertools import product
from urllib.parse import urljoin

import httpx

from pydantic import BaseModel

from ice.metrics.base import Metric
from ice.metrics.base import Sample
from ice.settings import settings
from ice.utils import map_async


class NubiaRequest(BaseModel):
    ref: str
    hyp: str


class NubiaResponse(BaseModel):
    nubia_score: float
    semantic_relation: float
    contradiction: float
    irrelevancy: float
    logical_agreement: float
    grammar_ref: float
    grammar_hyp: float


NUBIA_URL = urljoin(settings.OUGHT_INFERENCE_URL, "/nubia/")


async def _single_nubia(sample: Sample) -> list[NubiaResponse]:

    samples = list(product(sample.left, sample.right))

    async with httpx.AsyncClient(
        timeout=httpx.Timeout(connect=20, read=20, write=10, pool=30)
    ) as client:
        client.headers["x-api-key"] = settings.OUGHT_INFERENCE_API_KEY

        async def single_response(inputs: tuple[str, str]) -> NubiaResponse:
            hyp, ref = inputs
            response = await client.post(
                NUBIA_URL, json=NubiaRequest(ref=ref, hyp=hyp).dict()
            )
            return NubiaResponse.parse_obj(response.json())

        responses = await map_async(samples, single_response)

        return responses


class Nubia(Metric):
    name = "Nubia"

    async def compute(self, sample: list[Sample]) -> list[list[NubiaResponse]]:
        return await map_async(sample, _single_nubia)



import re

from collections.abc import Sequence
from typing import TypedDict

from structlog.stdlib import get_logger

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform

log = get_logger()


EXAMPLES: list[dict[str, str | ValueTransform | StopSentinel]] = [
    dict(
        question="What was the effect size?",
        selections=numbered_list(
            [
                """The WithAI score (88.87 ± 5.51) was significantly higher than the prelearning score  (75.73 ± 10.58, p < 0.01).""",
            ]
        ),
        reasoning="""We know that the WithAI score of 88.87 ± 5.51 was significantly higher than the prelearning score 75.73 ± 10.58, with p < 0.01). In order to answer the question fully, we should understand how the prelearning score and the WithAI scores were measured. We should also learn more about the intervention this difference applied to and whether other measures were included in the analysis.""",
        additional_questions=numbered_list(
            [
                "How was the prelearning score measured?",
                "How was the WithAI score measured?",
                "What intervention did the WithAI and prelearning scores apply to?",
                "In addition to the difference between the WithAI and prelearning scores, were there any other measures included in the analysis, and what were their effect sizes?",
            ]
        ),
        most_important_question="4: In addition to the difference between the WithAI and prelearning scores, were there any other measures included in the analysis, and what were their effect sizes?",
    ),
    dict(
        question="What was the design of the Ghana experiment?",
        selections=numbered_list(
            [
                """Of the six experiments, three are individual randomized trials with randomization at the household level within each village (India, Ethiopia and Pakistan) and three are clustered randomized trials, with randomization at both the village and household level (Ghana, Honduras, and Peru).""",
                """In the countries with clustered randomization, villages were randomly selected to be treatment or control villages, and then treatment households were randomly selected within the set of eligible households in treatment villages.""",
            ]
        ),
        reasoning="""We know that of the six experiments, three were individual randomized trials and three were clustered randomized trials. We also know that, in the countries with clustered randomization, villages were randomly selected to be treatment or control villages, and then treatment households were randomly selected within the set of eligible households in treatment villages. In order to answer the question fully, we should understand how the villages and households were selected in the individual randomized trials, and what the eligibility criteria were for households in the clustered randomized trials.""",
        additional_questions=numbered_list(
            [
                "How were villages and households selected in the Ghana experiment?",
                "What were the eligibility criteria for villages and households in the Ghana experiment?",
                "What were the treatment and control conditions in the clustered randomized trials, especially the Ghana experiment?",
                "What was the timing and duration of the Ghana experiment?",
            ]
        ),
        most_important_question="3: What were the treatment and control conditions in the clustered randomized trials, especially the Ghana experiment?",
    ),
]

PREFACE = "Instructions: For each question we have been asked to answer, we have been given some excerpts from a document containing the answer. Brainstorm 1-5 additional questions that would help us select more excerpts from the document, and then identify which of these questions would be most helpful to understand in order to be able to fully answer the original question."

EXAMPLE_TEMPLATE = """
Question we have been asked to answer: {question}

Excerpts from the document we have been given so far:

{selections}

What we know from these excerpts, and what additional information would help us answer the question: {reasoning}

1-5 additional questions to seek information on to answer the question we have been asked to answer:

{additional_questions}

Out of these questions, the question that would be most helpful to gather additional information about is question {most_important_question}""".strip()

EXAMPLE_SEPARATOR = "\n\n---\n\n"


prompt = "\n\n".join(
    (PREFACE, "\n\n---\n\n".join(format_multi(EXAMPLE_TEMPLATE, EXAMPLES)))
)


class RenderableAugmentQuestionExample(TypedDict):
    question: str
    selections: ValueTransform[Sequence[str]]
    reasoning: str
    additional_questions: ValueTransform[Sequence[str]]
    most_important_question: str


def make_augment_question_prompt(
    *,
    question: str,
    existing_selections: Sequence[str],
    examples: list[RenderableAugmentQuestionExample] | None = None,
) -> str:
    all_examples = (examples or EXAMPLES) + [
        dict(
            question=question,
            selections=numbered_list(existing_selections),
            reasoning=stop(""),
        )
    ]
    filled_examples = format_multi(EXAMPLE_TEMPLATE, all_examples)  # type: ignore[arg-type]
    prompt = "\n\n".join((PREFACE, EXAMPLE_SEPARATOR.join(filled_examples)))
    return prompt


def strip_enumeration_prefix(text: str) -> str:
    return re.sub(r"^\w*\s*\d+(\.|\))", "", text.strip()).strip()


def get_new_questions(completion: str) -> tuple[str, Sequence[str]]:
    # This is intentionally fragile to failure to follow the prompt format to first see how often/how it occurs
    _, remaining_part = completion.split("we have been asked to answer:")
    try:
        questions_list, remaining_part = remaining_part.split(
            "Out of these questions, the question that would be most helpful to gather additional information about is question "
        )
        num = int(remaining_part.split(":")[0].strip())
        questions = list(
            filter(
                None,
                (strip_enumeration_prefix(line) for line in questions_list.split("\n")),
            )
        )
        return questions[num - 1], questions
    except ValueError:
        log.warning("Unexpected response", completion=completion)
        questions = [
            strip_enumeration_prefix(line)
            for line in remaining_part.split("\n")
            if line and line[0].isnumeric()
        ]
        return questions[0], questions



from collections.abc import Sequence

from ice.recipes.meta.eval_text_classification import BinaryClassificationMetrics
from ice.recipes.meta.eval_text_classification import fuzzy_text_classification_metrics
from ice.recipes.meta.matching.match import match


async def eval_text_classification(
    candidates: Sequence[str],
    predictions: Sequence[bool],
    ground_truth: Sequence[str],
    scores: Sequence[float] | None = None,
) -> BinaryClassificationMetrics:
    return await fuzzy_text_classification_metrics(
        texts=candidates,
        predictions=predictions,
        ground_truth=ground_truth,
        scores=scores,
    )


async def eval_sequence_gen(
    question: str,
    ground_truth: Sequence[str],
    prediction: Sequence[str],
) -> tuple[bool, str]:
    question  # unused
    eval_detail: str
    correct: bool
    if len(ground_truth) < len(prediction):
        eval_detail, correct = "Too many items", False
    elif len(ground_truth) > len(prediction):
        eval_detail, correct = "Too few items", False
    elif len(ground_truth) == 1:
        eval_detail, correct = "One item, assuming correct", True
    else:
        eval_detail, correct = await match(ground_truth, prediction)
    return correct, eval_detail



from structlog.stdlib import get_logger

from ice.nn.bert_t5_t0_ensemble import BERT_T5_T0

log = get_logger()

# https://pypi.org/project/numerizer/ would be a more robust package for this however given the state of peotry it is not worth it
# Should be looked at again in the future


def extract_numbers(text: str) -> list[str]:
    words = text.split()

    set_number_str = {
        "zero",
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
        "ten",
        "teen",
        "twenty",
        "thirty",
        "forty",
        "fifty",
        "sixty",
        "seventy",
        "eighty",
        "ninety",
        "hundred",
        "thousand",
        "million",
        "billion",
        "trillion",
        "quadrillion",
        "quintillion",
    }
    number_strings = list(filter(lambda word: word.lower() in set_number_str, words))

    numbers_set = set("0123456789")
    number_strings += list(
        filter(lambda x: set(x).intersection(numbers_set) != set(), words)
    )

    # Remove parentheses
    remove_parentheses = (
        lambda s: s.replace("(", "")
        .replace(")", "")
        .replace("...", "")
        .replace("..", "")
    )
    number_strings = list(map(remove_parentheses, number_strings))

    # Remove "," or "." from the end of the number string
    for i, number in enumerate(number_strings):
        if number[-1] == "," or number[-1] == ".":
            number_strings[i] = number[:-1]

    return number_strings


def classify_example():

    abstract = """In this study we will examine the impact of the use of ..."""
    paragraph = """[..] The adherence rate is 88.2%."""
    numbers = extract_numbers(paragraph)
    category = "adherence"

    for number in numbers:
        classification = BERT_T5_T0(category, abstract, paragraph, number)
        log.info(f"{number}: {classification}")


if __name__ == "__main__":
    classify_example()



from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.value import numbered_list

PREAMBLE = "Convert the following model outputs into numbered lists:"

EXAMPLES = [
    dict(
        answer="The six experiments conducted in this paper were 1) an asset index experiment, 2) a consumption support experiment in Ethiopia and Peru, 3) a consumption support experiment in general, 4) an output production experiment, 5) an asset transfer experiment, and 6) a non-farm micro-enterprise experiment.",
        separated=numbered_list(
            [
                "An asset index experiment",
                "A consumption support experiment in Ethiopia and Peru",
                "A consumption support experiment in general",
                "An output production experiment",
                "An asset transfer experiment",
                "A non-farm micro-enterprise experiment",
            ]
        ),
    ),
    dict(
        answer="The one experiment conducted in this paper was a study of different quality control procedures for data collection. This study looked at how different quality control procedures may affect study outcomes, and the impact of quality control procedures on study outcomes.",
        separated=numbered_list(
            ["A study of different quality control procedures for data collection"]
        ),
    ),
    dict(
        answer=numbered_list(["a control group", "an intervention group"]),
        separated=numbered_list(["a control group", "an intervention group"]),
    ),
]

TEMPLATE = """Answer: {answer}

Convert to a numbered list:

{separated}"""


def make_quick_list_prompt(answer: str):
    examples = format_multi(
        TEMPLATE, EXAMPLES + [dict(answer=answer, separated=stop(""))]  # type: ignore[arg-type]
    )
    return "\n\n".join([PREAMBLE] + list(examples))



from functools import cache

from ice.agents.approval import ApprovalAgent
from ice.agents.augmented import AugmentedAgent
from ice.agents.base import Agent as Agent  # Explicit re-export
from ice.agents.cached import CachedAgent
from ice.agents.fake import FakeAgent
from ice.agents.human import HumanAgent
from ice.agents.openai import OpenAIAgent
from ice.agents.openai_reasoning import OpenAIReasoningAgent
from ice.agents.ought_inference import OughtInferenceAgent
from ice.agents.squad import SquadAgent
from ice.mode import Mode

# TODO: Come up with a more reasonable way to make agents optional/pluggable.
try:
    from ice.agents.tfew import TFew
except ImportError:

    class Tfew(Agent):
        def __init__(self, *args, **kwargs):
            ...


MACHINE_AGENTS = {
    "instruct": lambda: OpenAIAgent(),
    "instruct-reasoning": lambda: OpenAIReasoningAgent(),
    "instruct-reasoning-crowd": lambda: OpenAIReasoningAgent(num_workers=8),
    "curie": lambda: OpenAIAgent(model="curie"),
    "qasper": lambda: SquadAgent(),
    "mono-t5": lambda: OughtInferenceAgent(engine="mono-t5-base"),
    "adherence-tfew": lambda: TFew(
        origin_model_name="bigscience/T0_3B",
        lora_weights_path="ice/nn/weights/adherence_tfew_3B.pt",
    ),
    "adherence-tfew-multi": lambda: TFew(
        origin_model_name="bigscience/T0_3B",
        lora_weights_path="ice/nn/weights/adherence_tfew_multi_lite_3B.pt",
    ),
}


@cache
def _get_machine_agent(agent_name: str) -> Agent:
    return MACHINE_AGENTS[agent_name]()


def _get_augmented_agent(agent_name: str | None = None) -> AugmentedAgent:
    h_agent = agent_policy(mode="human", agent_name=agent_name)
    m_agent = agent_policy(mode="machine", agent_name=agent_name)
    return AugmentedAgent(human=h_agent, machine=m_agent)


def _get_approval_agent(agent_name: str | None = None) -> ApprovalAgent:
    h_agent = agent_policy(mode="human", agent_name=agent_name)
    m_agent = agent_policy(mode="machine", agent_name=agent_name)
    return ApprovalAgent(base_agent=m_agent, approval_agent=h_agent)


def agent_policy(mode: Mode, agent_name: str | None = None) -> Agent:
    if mode == "human":
        return HumanAgent()
    elif mode == "augmented":
        return _get_augmented_agent(agent_name)
    elif mode == "augmented-cached":
        return CachedAgent(_get_augmented_agent(agent_name))
    elif mode == "machine":
        return _get_machine_agent(agent_name or "instruct")
    elif mode == "machine-cached":
        return CachedAgent(
            _get_machine_agent(agent_name or "instruct"),
            cache_name=f"{(agent_name or 'instruct').replace('-', '_')}_cached",
        )
    elif mode == "fake" or mode == "test":
        return FakeAgent()
    elif mode == "approval":
        return _get_approval_agent(agent_name)
    else:
        raise ValueError(f"Unknown mode: {mode}")



from collections.abc import Sequence
from typing import ClassVar
from typing import Literal
from typing import Type

from pydantic import BaseModel
from pydantic import root_validator
from pydantic import validator

from ice.metrics.gold_standards import ParsedGoldStandardBase

NOT_YET_COLLECTED = "Gold standard not yet collected"


def _fix_keys(model_values: object, correct_key_mapping: dict[str, str]) -> object:
    if isinstance(model_values, dict):
        new_dict = dict()
        for key, value in model_values.items():
            if key in correct_key_mapping:
                new_dict[correct_key_mapping[key]] = value
            else:
                new_dict[key] = value
        return new_dict
    return model_values


def _not_yet_collected_is_none(value: object) -> object | None:
    if value == NOT_YET_COLLECTED:
        return None
    return value


def _maybe_dict_to_sequence(
    constructor: Type[BaseModel], maybe_dict: object
) -> object | Sequence[BaseModel]:
    if isinstance(maybe_dict, dict):
        return [
            constructor.parse_obj(vals | dict(name=k)) for k, vals in maybe_dict.items()
        ]
    return maybe_dict


class SampleSize(BaseModel):
    description: str | None = None
    reasoning: str | None = None
    n: int | Literal["Not mentioned"]
    quotes: Sequence[str]


class Reason(BaseModel):
    name: str
    description: str | None = None
    reasoning: str | None = None
    n: int | Literal["Not mentioned"]


class SampleSizeWithReasons(SampleSize):
    reasons: Literal["Not mentioned"] | Sequence[Reason] | None = None

    @validator("reasons", pre=True)
    def validate_reasons(cls, v):
        return _maybe_dict_to_sequence(Reason, v)


class Received(BaseModel):
    description: str | None = None
    quotes: Sequence[str]


class Analysis(SampleSize):
    name: str


class Arm(BaseModel):
    name: str
    allocated: None | Literal["Not mentioned"] | SampleSize
    received: None | Literal["Not mentioned"] | Received
    attrition: None | SampleSizeWithReasons
    analyzed: None | Sequence[Analysis]

    @validator("analyzed", pre=True)
    def validate_analyzed(cls, v):
        return _maybe_dict_to_sequence(Analysis, v)

    @root_validator(pre=True)
    def not_yet_collected_always_none(cls, values):
        if isinstance(values, dict):
            for key, value in values.items():
                if value == NOT_YET_COLLECTED:
                    values[key] = None
        return _fix_keys(
            values,
            {
                "Allocated to arm": "allocated",
                "Received allocated intervention": "received",
                "-> Attrition": "attrition",
                "Analysed": "analyzed",
            },
        )


class Enrolment(BaseModel):
    assessed: Literal["Not mentioned"] | SampleSize
    excluded: Literal["Not mentioned"] | SampleSizeWithReasons
    randomized: SampleSize

    @root_validator(pre=True)
    def fix_keys(cls, values):
        return _fix_keys(
            values,
            {
                "Assessed for eligibility": "assessed",
                "-> Excluded": "excluded",
                "Randomised": "randomized",
            },
        )


class Experiment(BaseModel):
    name: str
    description: str | None = None
    reasoning: str | None = None
    enrolment: Enrolment | None
    arms: Sequence[Arm] | None

    _enrolment_to_none = validator("enrolment", pre=True, allow_reuse=True)(
        _not_yet_collected_is_none
    )

    @validator("arms", pre=True)
    def validate_arms(cls, v):
        return _maybe_dict_to_sequence(Arm, v)

    @root_validator(pre=True)
    def fix_keys(cls, values):
        return _fix_keys(
            values, {"Enrolment": "enrolment", "assigning to arms and after": "arms"}
        )


class ConsortFlow(ParsedGoldStandardBase):
    question_short_name: ClassVar[str] = "consort_flow_v2"
    experiments: list[Experiment]

    @validator("experiments", pre=True)
    def validate_experiments(cls, v):
        return _maybe_dict_to_sequence(Experiment, v)



import enum
import string

from collections import Counter
from collections.abc import Mapping
from collections.abc import Sequence
from typing import final
from typing import TypeGuard

from structlog.stdlib import get_logger

from ice.formatter.transform import _Transform
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.positional import PositionalTransform
from ice.formatter.transform.value import ValueTransform

log = get_logger()

# Tools for building few-shot prompts out of examples

_formatter = string.Formatter()


class _NotNeededSentinel(enum.Enum):
    token = 0


_not_needed_sentinel = _NotNeededSentinel.token


@final
class StopSentinel(str):
    """Stop the string here."""


literal = str | int | float | StopSentinel

literal_or_transform = literal | _Transform

FormatValues = Sequence[Mapping[str, literal_or_transform]]

_StdLibFormatStringParses = list[tuple[str, str | None, str | None, str | None]]


def _is_partial(**fields: literal | _NotNeededSentinel):
    return any((isinstance(value, StopSentinel) for value in fields.values()))


def all_values_needed(
    examples: Sequence[Mapping[str, literal_or_transform | _NotNeededSentinel]]
) -> TypeGuard[Sequence[Mapping[str, literal_or_transform]]]:
    return all(
        (
            all(v is not _not_needed_sentinel for v in example.values())
            for example in examples
        )
    )


def _is_keyword_only(parses: _StdLibFormatStringParses) -> bool:
    return all(
        [
            field_name is None or field_name.isidentifier()
            for _, field_name, _, _ in parses
        ]
    )


def _parse(format_placeholder: str) -> _StdLibFormatStringParses:
    parses = list(_formatter.parse(format_placeholder))
    if not _is_keyword_only(parses):
        raise ValueError(
            f"placeholder must use only keyword placeholders: {format_placeholder}"
        )
    return parses


def _literal_brace_escape(literal_text: str) -> str:
    chars = Counter(literal_text)
    # For safety, ensuring these came from string.Formatter.parse
    assert chars.get("{", 0) + chars.get("}", 0) <= 1
    if not literal_text:
        return literal_text
    if literal_text[-1] == "}":
        return "".join((literal_text, "}"))
    if literal_text[-1] == "{":
        return "".join((literal_text, "{"))
    return literal_text


def _unparse(parses: _StdLibFormatStringParses) -> str:
    # Suprisingly this isn't in the stdlib
    parts: list[str] = []
    for literal_text, field_name, format_spec, conversion in parses:
        parts.append(_literal_brace_escape(literal_text))
        if field_name is not None:
            parts.extend(["{", field_name])
            if conversion is not None:
                parts.extend(["!", conversion])
            if format_spec is not None:
                parts.extend([":", format_spec])
            parts.append("}")
    return "".join(parts)


def _no_sentinels_remaining(
    concrete_values: dict[str, literal | _NotNeededSentinel]
) -> TypeGuard[dict[str, literal]]:
    return all(
        (value is not _not_needed_sentinel for value in concrete_values.values())
    )


def _format_truncate(
    format_placeholder: str,
    /,
    **concrete_values: literal | _NotNeededSentinel,
) -> str:
    if not _is_partial(**concrete_values):
        return format_placeholder
    parses = _parse(format_placeholder)
    keep: _StdLibFormatStringParses = []
    for parse in parses:
        _, field_name, _, _ = parse
        if field_name is None:
            keep.append(parse)
        else:
            keep.append(parse)
            if isinstance(concrete_values[field_name], StopSentinel):
                break
    needed_keys = set(
        (field_name for _, field_name, _, _ in keep if field_name is not None)
    )
    needed_concrete_values = {
        k: v for k, v in concrete_values.items() if k in needed_keys
    }
    if not _no_sentinels_remaining(needed_concrete_values):
        raise ValueError(
            f"Unfilled DependentTransform in partial example:\nFormat: {format_placeholder}\nNeeded Values: {needed_concrete_values}"
        )
    new_str = _unparse(keep).format_map(needed_concrete_values)
    return new_str


def _has_stop(concrete_values: Mapping[str, literal | _NotNeededSentinel]) -> bool:
    return any(isinstance(value, StopSentinel) for value in concrete_values.values())


def _format_single(
    format_placeholder: str,
    /,
    **concrete_values: literal | _NotNeededSentinel,
) -> tuple[str, bool]:
    if not _has_stop(concrete_values):
        formatted = format_placeholder.format_map(concrete_values)
        truncated = False
    else:
        formatted = _format_truncate(format_placeholder, **concrete_values)
        truncated = True
    return formatted, truncated


def _apply_transforms(
    cases: Sequence[Mapping[str, literal_or_transform]],
    total: int,
) -> Sequence[Mapping[str, literal | _NotNeededSentinel]]:
    ret_list: list[dict[str, literal | _NotNeededSentinel]] = []
    for position, case in enumerate(cases):
        ret_dict: dict[str, literal | _NotNeededSentinel] = {}
        for k, v in case.items():
            if not isinstance(v, _Transform):
                ret_dict[k] = v
            elif isinstance(v, ValueTransform):
                ret_dict[k] = v.transform()
            elif isinstance(v, PositionalTransform):
                ret_dict[k] = v.transform(position, total)
            elif isinstance(v, DependentTransform):
                if v.key() not in case:
                    ret_dict[k] = _not_needed_sentinel
                else:
                    dependent = case[v.key()]
                    ret_dict[k] = (
                        v.transform(dependent.value)
                        if isinstance(dependent, ValueTransform)
                        else v.transform(dependent)
                    )
        ret_list.append(ret_dict)
    return ret_list


_STOP_EARLY_WARNING = "StopSentinel used in non-final case; this may be a mistake."


def _format_multi(
    format_string: str,
    cases: Sequence[Mapping[str, literal_or_transform]],
    strip: bool,
) -> tuple[str, ...]:
    total = len(cases)
    transformed_examples = _apply_transforms(cases, total)
    filled_cases = list(
        map(
            lambda kwargs: _format_single(format_string, **kwargs),
            transformed_examples,
        )
    )
    ret_val, truncated = tuple(c[0].strip() if strip else c[0] for c in filled_cases), [
        c[1] for c in filled_cases
    ]
    if any(truncated[:-1]):
        log.warn(_STOP_EARLY_WARNING, format_string=format_string, cases=cases)
    return ret_val


def stop(value: str | int | float) -> StopSentinel:
    """Stop here; truncate the rest of the format string."""
    return StopSentinel(value)


def format_multi(
    format_str: str,
    cases: Sequence[Mapping[str, literal_or_transform]],
    shared: Mapping[str, literal_or_transform] | None = None,
    strip: bool = True,
) -> tuple[str, ...]:
    """
    str.Formatter on steroids.

    Repeatedly format a string based on the arguments passed in `cases`.
    Shared arguments belong in `shared`.

    Use Transforms to dynamically change the value based on other values,
    the index of the case, the number of cases, or another function.

    Stop early and truncate a case with a StopSentinel value.
    """
    shared_dict = {k: v for k, v in (shared or dict()).items()}
    return _format_multi(
        format_str, [shared_dict | case for case in cases], strip=strip
    )



from ice.recipes.single_prompt import SinglePrompt

QUESTION_SHORT_NAME = "placebo"

DEFAULT_ANSWER_CLASSIFICATION = "Placebo"

AI_PROMPT: str = "\nAssistant:"

HUMAN_PROMPT: str = "\nHuman:"

qa_prompt_template = f"""
{HUMAN_PROMPT} What is a placebo-controlled study?
{AI_PROMPT} A placebo-controlled study is a way of testing a treatment in which a separate control group receives a sham "placebo" treatment which is specifically designed to be indistinguishable from the real treatment but has no real effect.
{HUMAN_PROMPT} Here's the text of a paper I've been thinking about:

{{paper_text}}
{AI_PROMPT} I've read it thoroughly and can answer any question you may have about it.
{HUMAN_PROMPT} Was this a placebo-controlled study? Say "yes", "no", or "unclear".
{AI_PROMPT} Let's think step by step:"""


class PlaceboSimpleInstruct(SinglePrompt):
    agent_str = "instruct"
    max_tokens = 3500
    qa_prompt_template = qa_prompt_template
    question_short_name = QUESTION_SHORT_NAME
    default_answer_classification = DEFAULT_ANSWER_CLASSIFICATION



from collections.abc import Sequence

from structlog.stdlib import get_logger


log = get_logger()


CLUSTER_PREFIX = "Below are some answers to the question '{question}'"
CLUSTER_DIRECTIONS = """Cluster these answers by how they answer the question '{question}'. Write your answer in the form "Cluster N) [Answer]":

Cluster 1)"""

COUNT_DIRECTIONS = """How many answers are in each cluster? Write your answer in the form "Number of answers in Cluster N) [Number of answers in cluster N]":

Number of answers in cluster 1)"""


def candidates(
    answers: Sequence[str], prefix: str = "Answer", separator: str = "\n"
) -> str:
    return separator.join(
        (f"{prefix} {idx}) {answer.strip()}" for idx, answer in enumerate(answers, 1))
    )


def build_cluster_prompt(question: str, answers: Sequence[str]) -> str:
    prefix = CLUSTER_PREFIX.format(question=question)
    return "\n\n".join(
        (prefix, candidates(answers), CLUSTER_DIRECTIONS.format(question=question))
    )


def build_count_prompt(question: str, answers: Sequence[str], clusters: str) -> str:
    cluster_prompt = build_cluster_prompt(question, answers)
    return "\n\n".join((cluster_prompt + clusters.rstrip(), COUNT_DIRECTIONS))


def build_final_prompt(
    question: str, answers: Sequence[str], clusters: str, counts: str
):
    count_prompt = build_count_prompt(question, answers, clusters)
    return "\n\n".join(
        (
            count_prompt + counts.rstrip(),
            f"After this clustering procedure, we can conclude that the most popular answer to the question '{question}' is:\n\n",
        )
    )



from pathlib import Path

from ice.paper import Paper


def test_pdf():
    script_path = Path(__file__).parent
    paper = Paper.load(script_path / "../papers/keenan-2018.pdf")
    sentences = list(paper.sentences())
    assert (
        sentences[0]
        == "We hypothesized that mass distribution of a broad-spectrum antibiotic agent to preschool children would reduce mortality in areas of sub-Saharan Africa that are currently far from meeting the Sustainable Development Goals of the United Nations."
    )
    assert (
        sentences[-1]
        == "Disclosure forms provided by the authors are available with the full text of this article at NEJM.org."
    )


def test_txt():
    script_path = Path(__file__).parent
    paper = Paper.load(script_path / "../papers/keenan-2018-tiny.txt")
    sentences = list(paper.sentences())
    assert (
        sentences[0]
        == "In this cluster-randomized trial, we assigned communities in Malawi, Niger, and Tanzania to four twice-yearly mass distributions of either oral azithromycin (approxi- mately 20 mg per kilogram of body weight) or placebo."
    )
    assert (
        sentences[-1]
        == "The proportion of children whose census status was recorded as moved or unknown did not differ significantly between the groups (P=0.71 and P=0.36, respectively)."
    )



import json
import re

from collections.abc import Sequence
from itertools import islice

from pydantic import BaseModel
from pydantic import validator

from ice.cache import diskcache
from ice.paper import Paper
from ice.paper import Paragraph
from ice.paper import split_sentences
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard


TRAIN_PATH = "/code/datasets/qasper-train-v0.3.json"

VAL_PATH = "/code/datasets/qasper-dev-v0.3.json"


class RawAnswer(BaseModel):
    unanswerable: bool
    extractive_spans: list[str]
    yes_no: bool | None
    free_form_answer: str
    evidence: list[str]

    @validator("evidence")
    def transform_evidence(cls, v: list[str]) -> list[str]:
        return [_replace_references(para) for para in v]

    # TODO: Skip where any evidence starts with FLOAT_SELECTED

    def canonical_answer(self) -> str | None:
        if self.unanswerable:
            return "Not mentioned"
        answers = tuple(
            filter(None, (self.yes_no, self.free_form_answer, self.extractive_spans))
        )
        assert len(answers) == 1, "Expected exactly one answer type"
        answer = answers[0]
        if isinstance(answer, bool):
            return "yes" if answer else "no"
        elif isinstance(answer, str):
            return answer
        elif isinstance(answer, list):
            return "...".join(answer)
        else:
            raise TypeError(f"Unexpected answer type: {type(answer)}")


@diskcache()
def load_qasper_json(split: str) -> dict[str, dict]:
    path = TRAIN_PATH if split == "train" else VAL_PATH
    with open(path) as f:
        return json.load(f)


def _create_paper(paper_dict: dict, key: str) -> Paper:
    paper = Paper(
        document_id=key,
        paragraphs=[
            Paragraph(
                sectionType="abstract",
                sections=[],
                sentences=split_sentences(paper_dict["abstract"]),
            )
        ],
    )
    paragraphs: list[str] = sum([s["paragraphs"] for s in paper_dict["full_text"]], [])
    paragraphs = [p for p in paragraphs if p]
    for paragraph in paragraphs:
        paper.paragraphs.append(
            Paragraph(
                sectionType="main",
                sections=[],
                sentences=split_sentences(_replace_references(paragraph)),
            )
        )
    return paper


BIBREF_PATTERN = re.compile(r"BIBREF(\d+)")
TABFIGREF_PATTERN = re.compile(r"(TABREF|FIGREF)(\d+)")


def _match_to_subsequent_int_value(match: re.Match) -> str:
    return f"{int(match.groups()[-1]) + 1}"


def _replace_references(paragraph: str) -> str:
    unbibref = re.sub(
        BIBREF_PATTERN, lambda m: f"[{_match_to_subsequent_int_value(m)}]", paragraph
    )
    return re.sub(TABFIGREF_PATTERN, _match_to_subsequent_int_value, unbibref)


def _get_answers(question_dict: dict) -> tuple[str | None, list[str]]:
    answer_dict = question_dict["answers"][0][
        "answer"
    ]  # We'll only use the first answer

    answer = RawAnswer.parse_obj(answer_dict)

    return answer.canonical_answer(), answer.evidence


# class QasperSection(BaseModel):
#     section_name: str
#     paragraphs: Sequence[str]


# class QasperAnswer(BaseModel):
#     answer: RawAnswer


# class QasperQas(BaseModel):
#     question: str
#     question_id: str
#     answers: Sequence[QasperAnswer]


# class QasperPaperInfo(BaseModel):
#     title: str
#     abstract: str
#     full_text: Sequence[QasperSection]
#     qas: Sequence[QasperQas]


def generate_qasper_qas(split: str):
    qasper_info = load_qasper_json(split)
    for i, (id, paper_dict) in enumerate(qasper_info.items()):
        paper = _create_paper(paper_dict, id)
        for question_dict in paper_dict["qas"]:
            answer_str, evidence = _get_answers(question_dict)
            if answer_str is None:
                continue
            question = question_dict["question"]
            yield PaperQaGoldStandard(
                paper=paper,
                question=question,
                short_gold_answer=answer_str,
                gold_answer=answer_str,
                gold_support=evidence,
            )


def limited_generate_qasper_qas(split: str):
    LIMIT = 10
    yield from islice(generate_qasper_qas(split), LIMIT)


def qasper_support_func(
    document_id: str, GS: list[PaperQaGoldStandard]
) -> Sequence[PaperQaGoldStandard]:
    return [gs for gs in GS if gs.paper.document_id != document_id]



import httpx

from fvalues import F

from ice.recipe import recipe


def make_search_result_prompt(context: str, query: str, question: str) -> str:
    return F(
        f"""
Search results from Google for the query "{query}": "{context}"

Answer the following question, using the search results if helpful:

Question: "{question}"
Answer: "
"""
    ).strip()


def make_search_query_prompt(question: str) -> str:
    return F(
        f"""
You're trying to answer the question {question}. You get to type in a search query to Google, and then you'll be shown the results. What query do you want to search for?

Query: "
"""
    ).strip('" ')


async def search(query: str) -> dict:
    async with httpx.AsyncClient() as client:
        params = {"q": query, "hl": "en", "gl": "us", "api_key": "e29...7b4c"}
        response = await client.get("https://serpapi.com/search", params=params)
        return response.json()


def render_results(data: dict) -> str:
    if not data or not data.get("organic_results"):
        return "No results found"

    results = []
    for result in data["organic_results"]:
        title = result.get("title")
        link = result.get("link")
        snippet = result.get("snippet")
        if not title or not link or not snippet:
            continue
        results.append(F(f"{title}\n{link}\n{snippet}\n"))

    return F("\n").join(results)


async def choose_query(question: str) -> str:
    prompt = make_search_query_prompt(question)
    query = await recipe.agent().complete(prompt=prompt, stop='"')
    return query


async def answer_by_search(
    question: str = "Who is the president of the United States?",
) -> str:
    query = await choose_query(question)
    results = await search(query)
    results_str = render_results(results)
    prompt = make_search_result_prompt(results_str, query, question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(answer_by_search)



from ice.trace import TracedABC

Stop = str | list[str] | None


class Agent(TracedABC):
    label: str | None = None

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ) -> str:
        raise NotImplementedError

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: str | None = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], str | None]:
        raise NotImplementedError

    # Methods below may be deprecated in the future:

    async def relevance(
        self,
        *,
        context: str,
        question: str,
        verbose: bool = False,
        default: float | None = None,
    ) -> float:
        raise NotImplementedError

    async def predict(
        self, *, context: str, default: str = "", verbose: bool = False
    ) -> dict[str, float]:
        raise NotImplementedError



from collections.abc import Sequence

from ice.recipe import recipe
from ice.recipes.meta.matching.prompt import matching_prompt
from ice.recipes.meta.matching.prompt import MATCHING_STOP_SEQUENCES
from ice.recipes.meta.matching.prompt import reasoning_and_answer_from_completion


async def match(items_a: Sequence[str], items_b: Sequence[str]) -> tuple[str, bool]:
    assert len(items_a) == len(
        items_b
    ), "Matching intended only for sequences of equal length"
    prompt = matching_prompt(items_a, items_b)
    completion = await recipe.agent("alpha-current").complete(
        prompt=prompt,
        stop=MATCHING_STOP_SEQUENCES,
        max_tokens=600,
    )
    reasoning, answer = reasoning_and_answer_from_completion(completion)
    return reasoning, answer


recipe.main(match)



from collections.abc import Callable
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Generic
from typing import TypeVar

Reducer = Callable[[list[float]], float]


def reduce_scores_dict(
    scores: list[dict[str, float]], reducer: Reducer
) -> dict[str, float]:
    return {k: reducer([score[k] for score in scores]) for k in scores[0]}


@dataclass
class Sample:
    """
    Two lists of inputs to be compared via a metric.

    For example, left could be a list of hypothesis generations
    and right some reference (gold standard) examples.

    Or left could be human-generated examples and right could be
    GPT-generated examples.
    """

    left: Sequence[str]
    right: Sequence[str]

    def __post_init__(self):
        if not self.left or not self.right:
            raise ValueError("Each sample must have at least one element")

    def identical_sample(self):
        return set(self.left) == set(self.right)


RetVal = TypeVar("RetVal")


class Metric(Generic[RetVal]):
    async def compute(self, samples: list[Sample]) -> list[RetVal]:
        raise NotImplementedError



import pytest

from ice.utils import nsmallest_async


async def cmp(x: int, y: int) -> int:
    return x - y


@pytest.mark.anyio
async def test_nsmallest_async():
    xs = [7, 8, 4, 3, 1, 6, 2, 0, 9, 5]
    assert await nsmallest_async(3, xs, cmp) == [0, 1, 2]
    assert xs == [7, 8, 4, 3, 1, 6, 2, 0, 9, 5]

    assert await nsmallest_async(1, [], cmp) == []
    assert await nsmallest_async(0, [1], cmp) == []
    assert await nsmallest_async(1, [2, 1, 3], cmp) == [1]
    assert await nsmallest_async(-1, [2, 1, 3], cmp) == []
    assert await nsmallest_async(4, [2, 1, 3], cmp) == [1, 2, 3]
    assert await nsmallest_async(1, [1], cmp) == [1]
    assert await nsmallest_async(3, list(range(10)), cmp) == [0, 1, 2]



from fvalues import F

from ice.recipe import recipe


def make_verification_prompt(question: str, answer: str) -> str:
    return F(
        f"""Consider this question: "{question}"

Potential answer: "{answer}"

Q: Is the potential answer above correct? Say "A: Yes" or "A: No".
A:"""
    )


async def verify_answer(question: str, answer: str) -> float:
    prompt = make_verification_prompt(question=question, answer=answer)
    choice_probs, _ = await recipe.agent().classify(
        prompt=prompt, choices=(" Yes", " No")
    )
    return choice_probs.get(" Yes", 0)


recipe.main(verify_answer)



from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_full_document_id
from ice.paper import Paper
from ice.recipe import Recipe
from ice.utils import truncate_by_tokens


def get_paper_text(paper: Paper) -> str:
    return "\n\n".join(str(p) for p in paper.paragraphs)


def create_recipe_result(
    paper_id: str,
    experiment: str,
    answer: str,
    question_short_name: str,
    default_answer_classification: str | None,
) -> RecipeResult:
    return RecipeResult(
        document_id=paper_id,
        question_short_name=question_short_name,
        experiment=experiment,
        classifications=[
            default_answer_classification,
            default_answer_classification,
        ],
        answer=answer,
        result=answer,
        excerpts=[],
    )


class SinglePrompt(Recipe):
    agent_str: str
    max_tokens: int
    qa_prompt_template: str
    question_short_name: str
    default_answer_classification: str | None

    async def run(self, paper: Paper):

        # Get the full paper text and truncate it
        full_paper_text = get_paper_text(paper)
        paper_text = truncate_by_tokens(full_paper_text, max_tokens=self.max_tokens)

        # Generate the QA prompt
        qa_prompt = self.qa_prompt_template.format(paper_text=paper_text)

        # Ask the agent to answer the prompt
        answer = await self.agent(self.agent_str).complete(
            prompt=qa_prompt, max_tokens=300
        )

        # Save results for each experiment
        paper_id = get_full_document_id(paper.document_id)
        experiments: list[str] = list_experiments(document_id=paper_id)
        recipe_results = [
            create_recipe_result(
                paper_id,
                experiment,
                answer,
                self.question_short_name,
                self.default_answer_classification,
            )
            for experiment in experiments
        ]
        self.maybe_add_to_results(recipe_results)

        return answer



import asyncio
import importlib
import sys

from abc import abstractmethod
from collections.abc import Awaitable
from collections.abc import Callable
from functools import wraps
from inspect import iscoroutinefunction
from pathlib import Path
from traceback import print_exc
from typing import final
from typing import Generic
from typing import no_type_check
from typing import TypeGuard
from typing import TypeVar

import defopt
import pandas as pd

from merge_args import merge_args
from pydantic import BaseSettings
from structlog.stdlib import get_logger

from ice.agent import Agent
from ice.agent import agent_policy
from ice.environment import env
from ice.evaluation.evaluate_recipe_result import EvaluatedRecipeResult
from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.evaluation.evaluation_report import EvaluationReport
from ice.mode import Mode
from ice.paper import Paper
from ice.trace import enable_trace
from ice.trace import trace
from ice.trace import TracedABC
from ice.utils import map_async

RecipeSettings = TypeVar("RecipeSettings", bound=BaseSettings)

log = get_logger()


def is_list_of_recipe_result(value: object) -> TypeGuard[list[RecipeResult]]:
    return isinstance(value, list) and all(
        isinstance(item, RecipeResult) for item in value
    )


class Recipe(TracedABC, Generic[RecipeSettings]):
    defaults: Callable[["Recipe"], RecipeSettings] = lambda self: BaseSettings()

    def __init__(self, mode: Mode = "machine", settings: RecipeSettings | None = None):
        self.mode = mode
        self.s = settings or self.defaults()  # type: ignore[call-arg,misc]
        self.results: list[RecipeResult] = []

    @classmethod
    def slug(cls) -> str:
        """A unique identifier for this recipe, which does not change when the recipe is updated."""
        return cls.__name__.lower()

    @no_type_check
    @abstractmethod
    async def run(self, **kwargs):
        raise NotImplementedError

    @final
    def maybe_add_to_results(self, results: list[RecipeResult] | object):
        if is_list_of_recipe_result(results):
            self.results.extend(results)

    def to_json(self, results: list[RecipeResult] | object) -> list[dict]:
        """Convert results to objects that can be serialized to JSON."""
        if is_list_of_recipe_result(results):
            return [result.dict() for result in results]
        raise NotImplementedError

    async def evaluation_report(self) -> EvaluationReport:
        return EvaluationReport(
            technique_name=str(self),
            results=await map_async(
                self.results, EvaluatedRecipeResult.from_recipe_result
            ),
        )

    def agent(self, agent_name: str | None = None) -> Agent:
        return agent_policy(mode=self.mode, agent_name=agent_name)

    def max_concurrency(self) -> int:
        return 10 if self.mode == "machine" else 1

    def __str__(self) -> str:
        return self.__class__.__name__


FunctionBasedRecipe = Callable[..., Awaitable]


def function_recipe_from_path(path: str) -> FunctionBasedRecipe:
    # from https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    file_path, _, recipe = path.partition(":")
    module_name = file_path.replace("/", ".").replace(".py", "")
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    assert spec
    module = importlib.util.module_from_spec(spec)
    assert spec.loader
    spec.loader.exec_module(module)
    return vars(module)[recipe]


class RecipeHelper:
    def __init__(self):
        self._mode: Mode | None = "machine"
        self.all_recipes: list[FunctionBasedRecipe] = []

    def main(self, main: FunctionBasedRecipe):
        if not iscoroutinefunction(main):
            raise TypeError("recipe.main must be given an async function")

        # Trace all globals defined in main's module.
        try:
            g = main.__globals__
        except AttributeError:
            # Perhaps this is a functools.partial
            g = main.func.__globals__  # type: ignore[attr-defined]
        for name, value in g.items():
            if getattr(value, "__module__", None) == main.__module__:
                g[name] = trace(value)

        traced_main = trace(main)
        self.all_recipes.append(traced_main)

        if main.__module__ != "__main__":
            return

        # The frontend shows everything under the first traced root.
        # TODO: Once main.py is gone, change the frontend and get rid of this wrapper.
        @trace
        @wraps(main)
        async def hidden_wrapper(*args, **kwargs):
            try:
                result = await traced_main(*args, **kwargs)
            except NameError:
                print_exc()
                print(
                    "\nReminder: recipe.main should be at the bottom of the file",
                    file=sys.stderr,
                )
                sys.exit(1)

            env().print(result, format_markdown=False)
            return result

        # A traced function cannot be called until the event loop is running.
        @wraps(main)
        async def untraced_wrapper(*args, **kwargs):
            return await hidden_wrapper(*args, **kwargs)

        @merge_args(main)
        def cli(
            *args,
            mode: Mode = "machine",
            trace: bool = True,
            **kwargs,
        ):
            self._mode = mode
            if trace:
                enable_trace()
            asyncio.run(untraced_wrapper(*args, **kwargs))

        defopt.run(
            cli,
            cli_options="all",
            short={},
            parsers={
                Paper: lambda path: Paper.load(Path(path)),
                FunctionBasedRecipe: function_recipe_from_path,
                pd.DataFrame: lambda path: pd.read_csv(path),
            },
        )

    def agent(self, agent_name: str | None = None) -> Agent:
        assert self._mode
        return agent_policy(self._mode, agent_name)


recipe = RecipeHelper()



from collections.abc import Sequence
from typing import cast

from ice.apis.openai import openai_complete
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import plural_transform
from ice.paper import Paper
from ice.recipe import Recipe
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.golds import get_ea_gs
from ice.recipes.experiments_and_arms.num_utils import strip_enumeration_prefix
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    CAN_WE_NAME_EXPERIMENTS_BEST_CHOICE,
)
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    CAN_WE_NAME_EXPERIMENTS_CHOICES,
)
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    CAN_WE_NAME_EXPERIMENTS_REASONING_STOP,
)
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    get_can_we_name_experiments_helpfulness,
)
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    get_can_we_name_experiments_reasoning,
)
from ice.recipes.experiments_and_arms.prompts.can_name_exps import (
    make_can_we_name_experiments_prompt,
)
from ice.recipes.experiments_and_arms.prompts.name_exps import get_name_exps_reasoning
from ice.recipes.experiments_and_arms.prompts.name_exps import make_name_exps_from_count
from ice.recipes.experiments_and_arms.prompts.name_exps import (
    NAME_EXPERIMENTS_REASONING_STOP,
)
from ice.recipes.experiments_and_arms.prompts.passages_to_keep import (
    keep_most_helpful_paragraphs,
)
from ice.recipes.experiments_and_arms.prompts.quick_list import make_quick_list_prompt
from ice.recipes.experiments_and_arms.recipes.best_passages import (
    rate_helpfulness_with_reasoning,
)
from ice.recipes.experiments_and_arms.recipes.consensus import best_answer_by_consensus
from ice.recipes.experiments_and_arms.recipes.count_experiments import count_experiments
from ice.recipes.experiments_and_arms.recipes.reason_select_and_answer import (
    answer_with_best_reasoning,
)
from ice.recipes.experiments_and_arms.types import PassageWithReasoning
from ice.trace import Recorder
from ice.trace import recorder
from ice.trace import trace


async def first(exps: Sequence[PassageWithReasoning[str]]) -> PassageWithReasoning[str]:
    return exps[0]


def make_reduce_to_best_answer(num_experiments: int):
    async def reduce_to_best_answer(
        candidates: Sequence[PassageWithReasoning[str]],
    ) -> PassageWithReasoning[str]:
        answers = [c.final_answer for c in candidates if c.final_answer]
        question = f"""What {plural_transform("key", "was", "were").transform(num_experiments)} the {CountWord("key").transform(num_experiments)} experiment{plural_transform("key", "", "s").transform(num_experiments)} conducted in this paper?"""
        return PassageWithReasoning(
            passage=candidates[0].passage,
            reasoning="",
            helpfulness="",
            final_answer=await (
                # best_answer_by_clustering(question=question, candidates=answers)
                best_answer_by_consensus(question=question, candidates=answers)
            ),
        )

    return reduce_to_best_answer


@trace
async def best_paras_for_naming_experiments(paper: Paper):
    experiment_count, _ = await count_experiments(paper)
    assert experiment_count.final_answer is not None
    paragraphs = [str(p) for p in paper.nonempty_paragraphs()]
    passages_by_relevance = await rate_helpfulness_with_reasoning(
        paragraphs,
        make_can_we_name_experiments_prompt(experiment_count.final_answer),
        CAN_WE_NAME_EXPERIMENTS_CHOICES,
        CAN_WE_NAME_EXPERIMENTS_BEST_CHOICE,
        reasoning_stop=CAN_WE_NAME_EXPERIMENTS_REASONING_STOP,
        get_reasoning=get_can_we_name_experiments_reasoning,
        get_helpfulness=get_can_we_name_experiments_helpfulness,
        num_shots=3,
        passages_per_prompt=4,
        step=1,
    )

    paragraphs_to_keep = await keep_most_helpful_paragraphs(passages_by_relevance)
    return [p in paragraphs_to_keep for p in paragraphs], paragraphs


@trace
async def name_experiments(
    paper: Paper, record: Recorder = recorder
) -> tuple[Sequence[str], Sequence[str], Sequence[str], Sequence[str]]:
    """What were the experiments conducted in this paper?

    Args:
        paper (Paper): The paper in question.
        record (Recorder, optional): (recorder for tracing). Defaults to recorder.

    Returns:
        tuple[Sequence[str], Sequence[str]]: The gold standard experiments and the generated experiments.
    """
    gs = get_ea_gs(paper.document_id)
    if gs and gs.parsed_answer:
        gs_names = [exp.name for exp in gs.parsed_answer.experiments]
    else:
        gs_names = []
    experiment_count, _ = await count_experiments(paper)
    assert experiment_count.final_answer is not None

    paragraphs = paper.nonempty_paragraphs()
    passages_by_relevance = await rate_helpfulness_with_reasoning(
        [str(p) for p in paragraphs],
        make_can_we_name_experiments_prompt(experiment_count.final_answer),
        CAN_WE_NAME_EXPERIMENTS_CHOICES,
        CAN_WE_NAME_EXPERIMENTS_BEST_CHOICE,
        reasoning_stop=CAN_WE_NAME_EXPERIMENTS_REASONING_STOP,
        get_reasoning=get_can_we_name_experiments_reasoning,
        get_helpfulness=get_can_we_name_experiments_helpfulness,
        num_shots=3,
        passages_per_prompt=4,
        step=1,
    )

    paragraphs_to_keep = await keep_most_helpful_paragraphs(passages_by_relevance)

    experiment_names = await answer_with_best_reasoning(
        num_samples=10,  # TODO: Better sampling here
        selector=make_reduce_to_best_answer(
            num_experiments=experiment_count.final_answer
        ),
        # selector=first,
        texts=paragraphs_to_keep,
        # texts=count_paras,
        num_examples=2,
        reasoning_temperature=0.4,
        reasoning_stop=NAME_EXPERIMENTS_REASONING_STOP,
        prompt_func=make_name_exps_from_count(experiment_count.final_answer),
        get_reasoning=get_name_exps_reasoning,
        get_helpfulness=None,
        final_answer_processor=lambda resp: cast(str, resp["choices"][0]["text"]),
    )

    standardized_answer = await convert_answer_to_standardized_format(
        experiment_names.final_answer
    )

    assert experiment_names.final_answer is not None
    return (
        gs_names,
        [
            strip_enumeration_prefix(exp_name)
            for exp_name in standardized_answer.split("\n")
            if exp_name.strip()
        ]
        if standardized_answer
        else [],
        paragraphs_to_keep,
        [str(p) for p in paragraphs],
    )


@trace
async def convert_answer_to_standardized_format(answer: str) -> str:
    standardized_answer: str = (
        await openai_complete(
            prompt=make_quick_list_prompt(answer),
            stop="\n\nAnswer:",
        )
    )["choices"][0]["text"]
    return standardized_answer


class NameExperiments(Recipe):
    async def run(self, paper: Paper):
        return await name_experiments(paper)


recipe.main(name_experiments)



from structlog.stdlib import get_logger

from ice.paper import Paper
from ice.recipe import Recipe
from ice.recipes.comparisons_qa import ComparisonsQA
from ice.utils import map_async

Arm = str

log = get_logger()


class ExperimentArms(Recipe):
    do_not_test = True

    async def get_arms(self, paper: Paper, experiment: str) -> list[Arm]:
        qa_recipe = ComparisonsQA(mode=self.mode)
        arms_result = await qa_recipe.run(
            paper=paper,
            question_short=f"""What were the trial arms (subgroups of participants) of the experiment specifically described as {experiment}?""",
            question_long=f"""What were the trial arms (subgroups of participants) of the experiment specifically described as {experiment}? List one per line.""",
            answer_prefix="Answer: The trial arms were:\n-",
        )
        arms = [line.strip("- ") for line in arms_result.split("\n")]
        return arms

    async def describe_arm(
        self, paper: Paper, arms: list[Arm], arm_to_describe: Arm, experiment: str
    ) -> str:
        qa_recipe = ComparisonsQA(mode=self.mode)
        if len(arms) == 1:
            qa_result = await qa_recipe.run(
                paper=paper,
                question_short=f"""What was the setup of the "{arm_to_describe}" trial arm in the experiment described as {experiment}?""",
                question_long=f"""What was the setup of the "{arm_to_describe}" trial arm in the experiment described as {experiment}? Describe the intervention in a few sentences.""",
                answer_prefix="Answer:",
                num_paragraphs=2,
            )
        else:
            arm_strings = ", ".join(arms)
            qa_result = await qa_recipe.run(
                paper=paper,
                question_short=f"""What was the setup of the "{arm_to_describe}" trial arm in the experiment described as {experiment}?""",
                question_long=f"""In the experiment described as {experiment} there were {len(arms)} trial arms, {arm_strings}. What was the setup of the "{arm_to_describe}" trial arm? Describe the intervention in a few sentences.""",
                answer_prefix="Answer:",
                num_paragraphs=2,
            )
        return qa_result

    async def get_arm_descriptions(
        self, paper: Paper, arms: list[Arm], experiment: str
    ) -> list[str]:
        arm_descriptions = await map_async(
            arms,
            lambda arm_to_describe: self.describe_arm(
                paper, arms, arm_to_describe, experiment
            ),
            max_concurrency=self.max_concurrency(),
        )
        return arm_descriptions

    async def run(self, paper: Paper, experiment):
        arms = await self.get_arms(paper, experiment)

        arm_descriptions = await self.get_arm_descriptions(paper, arms, experiment)

        return arms, arm_descriptions



import random

from faker import Faker

from ice.agents.base import Agent
from ice.agents.base import Stop

random.seed(0)
Faker.seed(0)


class FakeAgent(Agent):
    def __init__(self):
        self.fake = Faker()

    async def relevance(self, *, question, context, verbose=False, default=None):
        return random.random()

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ):
        return self.fake.sentence()

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: str | None = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], str | None]:
        probs = [random.random() for _ in choices]
        total = sum(probs)
        probs = [p / total for p in probs]
        distribution = dict(zip(choices, probs))
        return distribution, None

    async def predict(
        self, *, context: str, default: str = "", verbose: bool = False
    ) -> dict[str, float]:
        if default:
            words = [default]
        else:
            words = []
        words += [self.fake.word() for _ in range(random.randint(1, 5))]
        return {word: random.random() for word in words}



from dataclasses import dataclass
from typing import Protocol


@dataclass
class Example:
    question: str | None
    texts: list[str]
    compressed: list[str]


class Compress(Protocol):
    async def __call__(
        self, question: str | None, texts: list[str], examples: list[Example]
    ) -> list[str]:
        pass



from ice.recipe import recipe
from ice.recipes.primer.answer_by_dispatch.prompt import *


async def answer_by_dispatch(question: str = "How many people live in Germany?"):
    prompt = make_action_selection_prompt(question)
    choices = tuple(str(i) for i in range(1, 6))
    probs, _ = await recipe.agent().classify(prompt=prompt, choices=choices)
    return list(zip(probs.items(), [a.name for a in action_types]))


recipe.main(answer_by_dispatch)



from collections.abc import Sequence
from datetime import datetime

from ice.metrics.base import reduce_scores_dict
from ice.metrics.base import Sample
from ice.metrics.rouge import Rouge
from ice.metrics.rouge import RougeResult
from ice.settings import OUGHT_ICE_DIR

CSVS_PATH = OUGHT_ICE_DIR / "evaluation_csvs/"


def _partial_confusion_matrix(
    actuals: list[bool], predictions: list[bool]
) -> tuple[int, int, int]:
    assert len(actuals) == len(predictions)
    true_positives = 0
    false_positives = 0
    false_negatives = 0
    for actual, prediction in zip(actuals, predictions):
        if actual and prediction:
            true_positives += 1
        elif not actual and prediction:
            false_positives += 1
        elif actual and not prediction:
            false_negatives += 1
    return true_positives, false_positives, false_negatives


def precision_score(actuals: list[bool], predictions: list[bool]) -> float:
    true_positives, false_positives, _ = _partial_confusion_matrix(actuals, predictions)
    return (
        true_positives / (true_positives + false_positives) if true_positives else 0.0
    )


def recall_score(actuals: list[bool], predictions: list[bool]) -> float:
    true_positives, _, false_negatives = _partial_confusion_matrix(actuals, predictions)
    return (
        true_positives / (true_positives + false_negatives) if true_positives else 0.0
    )


def mean(values: list[float]) -> float:
    return sum(values) / len(values) if values else 0.0


def top(values: list[float]) -> float:
    return max(values) if values else 0.0


async def rouge_compare(hyp: Sequence[str], ref: Sequence[str]) -> RougeResult:
    sample = Sample(left=hyp, right=ref)
    rouge = Rouge()
    scores = await rouge.compute([sample])
    assert len(scores) == 1
    return scores[0]


def summarize_scores(scores: list[RougeResult]) -> RougeResult | None:
    if not scores:
        return None
    summary: dict[str, dict[str, float]] = dict()
    for metric in scores[0].dict(by_alias=True):
        summary[metric] = reduce_scores_dict(
            [score.dict(by_alias=True)[metric] for score in scores], mean
        )

    return RougeResult.parse_obj(summary)


start_time = datetime.now().astimezone()



from fvalues import F

from ice.recipe import recipe
from ice.recipes.primer.verify.utils import *


def make_verification_prompt(question: str, steps: list[str]) -> str:
    return F(
        f"""Consider this question: "{question}"

Here are the first few steps of an answer:

{render_steps(steps)}

Q: Is step {len(steps)} correct, assuming that the previous steps are correct? Say "A: Yes" or "A: No".
A:"""
    )


async def check_step(question: str, steps: list[str]) -> float:
    """
    Return the probability that the step is correct
    """
    prompt = make_verification_prompt(question=question, steps=steps)
    answer_probs, _ = await recipe.agent().classify(
        prompt=prompt, choices=(" Yes", " No")
    )
    return answer_probs.get(" Yes", 0.0)


async def verify_answer(
    question: str = DEFAULT_QUESTION, steps: list[str] = DEFAULT_STEPS
):
    return await check_step(question=question, steps=steps)


recipe.main(verify_answer)



import json

from collections.abc import Sequence

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.paper import Paper
from ice.recipe import Recipe
from ice.recipes.experiments_and_arms.golds import get_ea_gs
from ice.recipes.experiments_and_arms.recipes.name_arms import name_arms
from ice.recipes.experiments_and_arms.recipes.name_experiments import name_experiments
from ice.recipes.experiments_and_arms.recipes.quick_evaluate import quick_evaluate
from ice.recipes.experiments_and_arms.types import Arm
from ice.recipes.experiments_and_arms.types import Experiment
from ice.recipes.experiments_and_arms.types import ExperimentsArms
from ice.trace import recorder
from ice.trace import trace
from ice.utils import map_async


@trace
async def experiments_and_arms(
    paper: Paper, record=recorder
) -> tuple[str, str, ExperimentsArms, ExperimentsArms]:
    """What were the experiments performed in this paper, and for each experiment, what were the trial arms?

    Args:
        paper (Paper): Paper to evaluate.
        record (Recorder, optional): (recorder for tracing). Defaults to recorder.

    Returns:
        tuple[str, str, ExperimentsArms, ExperimentsArms]: The quick-eval grade, the explanation for that grade, the gold standard answer, and the generated answer.
    """
    gs = get_ea_gs(paper.document_id)
    _, exps, _, _ = await name_experiments(paper)

    async def run_arms(experiment_in_question: str) -> Sequence[str]:
        return await name_arms(
            paper=paper, experiments=exps, experiment_in_question=experiment_in_question
        )

    arms_by_exp = await map_async(exps, run_arms)
    result = ExperimentsArms(
        experiments=[
            Experiment(
                name=exp,
                description="",
                arms=[Arm(name=a, description="") for a in arm],
            )
            for exp, arm in zip(exps, arms_by_exp)
        ]
    )
    gs_answer = (
        gs.parsed_answer if gs and gs.parsed_answer else ExperimentsArms(experiments=[])
    )

    evaluation, grade = await quick_evaluate(gs_answer, result)

    return grade, evaluation, gs_answer, result


class ExperimentsAndArms(Recipe):
    async def run(self, paper: Paper):
        """Wrapped experiments and arms recipe for evaluation script.

        Args:
            paper (Paper): Paper to evaluate.
        """
        grade, evaluation, _, result = await experiments_and_arms(paper)
        self.maybe_add_to_results(
            [
                RecipeResult(
                    question_short_name="experiments_arms",
                    document_id=paper.document_id,
                    experiment="",
                    excerpts=[],
                    answer=json.dumps(
                        result.dict()
                        | dict(autoeval_grade=grade, autoeval_reasoning=evaluation)
                    ),
                    result=result,
                )
            ]
        )



from os import environ
from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import BaseSettings

if TYPE_CHECKING:
    AnyHttpUrl = str

else:
    from pydantic import AnyHttpUrl


class Settings(BaseSettings):
    OPENAI_API_KEY: str = ""
    OPENAI_ORG_ID: str = ""
    OUGHT_INFERENCE_API_KEY: str = ""
    OUGHT_INFERENCE_URL: AnyHttpUrl = "https://prod.elicit.org"
    ELICIT_AUTH_TOKEN: str = ""
    GOLD_STANDARDS_CSV_PATH: Path = (
        Path(__file__).parent.parent / "gold_standards/gold_standards.csv"
    )
    GS_QUOTE_FOUND_THRESHOLD: float = 0.75
    OUGHT_ICE_HOST: str = "0.0.0.0"
    OUGHT_ICE_PORT: int = 8935
    OUGHT_ICE_AUTO_SERVER: bool = True
    OUGHT_ICE_AUTO_BROWSER: bool = True
    PAPER_DIR: Path = Path(__file__).parent.parent / "papers"


OUGHT_ICE_DIR = Path(environ.get("OUGHT_ICE_DIR", Path.home() / ".ought-ice"))

_env_path = OUGHT_ICE_DIR / ".env"
settings = Settings(
    _env_file=_env_path if _env_path.exists() else None, _env_file_encoding="utf-8"
)

CACHE_DIR = OUGHT_ICE_DIR / "cache"
CACHE_DIR.mkdir(parents=True, exist_ok=True)


def server_url() -> str:
    return f"http://{settings.OUGHT_ICE_HOST}:{settings.OUGHT_ICE_PORT}"



import re

from collections.abc import Iterator
from pathlib import Path
from typing import Annotated
from typing import Literal
from typing import Optional

import nltk
import requests

from nltk.tokenize import sent_tokenize
from pydantic import BaseModel
from pydantic import Field
from structlog.stdlib import get_logger

from ice.cache import diskcache
from ice.environment import env
from ice.settings import OUGHT_ICE_DIR

log = get_logger()

PDF_PARSER_URL = "https://test.elicit.org/elicit-previews/james/oug-3083-support-parsing-arbitrary-pdfs-using/parse_pdf"

SectionType = Literal["abstract", "main", "back"]

nltk.download("punkt", quiet=True)


def get_full_document_id(document_id: str) -> str:
    match document_id:
        case "abebe-2018-tiny.txt":
            return "abebe-2018.pdf"
        case "keenan-2018-tiny.txt":
            return "keenan-2018.pdf"
        case _:
            return document_id


def get_paper_paths(paper_dir: Path | None = None) -> list[Path]:
    if paper_dir is None:
        script_path = Path(__file__).parent.parent
        paper_dir = script_path / "papers/"
    paper_paths = [f for f in paper_dir.iterdir() if f.suffix in (".pdf", ".txt")]
    paper_paths = sorted(paper_paths)
    return paper_paths


def starts_with_number(s: str) -> bool:
    return re.match(r"^\d+", s) is not None


def is_likely_section_title(text: str):
    return (starts_with_number(text) and len(text) < 200) or text == "Abstract"


def split_sentences(text: str) -> list[str]:
    return sent_tokenize(text)


def section_title_words(title: str) -> str:
    if not starts_with_number(title):
        return title
    return " ".join(title.split(" ")[1:])


def section_title_number(title: str) -> str:
    if not starts_with_number(title):
        return ""
    return title.split(" ")[0]


def parse_txt(file: Path) -> list[dict]:
    body = []
    with open(file) as text_file:
        text = text_file.read()
        paragraphs = text.split("\n\n")
        current_section = "Introduction"
        for paragraph in paragraphs:
            if is_likely_section_title(paragraph):
                current_section = paragraph
            else:
                sentences = split_sentences(paragraph)
                body.append(
                    {
                        "sentences": sentences,
                        "sections": [
                            {
                                "title": section_title_words(current_section),
                                "number": section_title_number(current_section),
                            }
                        ],
                        "sectionType": "abstract"
                        if current_section == "Abstract"
                        else "main",
                    }
                )
    return body


def save_pdf_text(paper_body: list[dict], file_name: str):
    """
    Save pdf text to help with debugging.
    """
    paper_txt_dir = OUGHT_ICE_DIR / "paper_txt"
    paper_txt_dir.mkdir(parents=True, exist_ok=True)
    open(paper_txt_dir / f"{file_name}.txt", "w").write(
        "\n\n".join(" ".join(paragraph["sentences"]) for paragraph in paper_body)
    )


@diskcache()
def parse_pdf(file: Path) -> list[dict]:
    with env().spinner(f"[bold green] Parsing {file.name}"):
        files = {"pdf": open(file, "rb")}
        r = requests.post(PDF_PARSER_URL, files=files)
        body = r.json()
        save_pdf_text(body, file.name)
    env().print(f"Parsed {file.name}.")
    return body


class Section(BaseModel):
    title: str
    number: Optional[str] = None


class Paragraph(BaseModel):
    sentences: list[str]
    sections: list[Section]
    section_type: Annotated[SectionType, Field(alias="sectionType")]

    def is_body_paragraph(self):
        return self.section_type in ("abstract", "main")

    def is_empty(self):
        return str(self).strip() == ""

    def __str__(self):
        return " ".join(self.sentences)

    def __hash__(self):
        return hash(str(self))


class Paper(BaseModel):
    paragraphs: list[Paragraph]
    document_id: str = "unknown"

    @classmethod
    def load(cls, file: Path) -> "Paper":
        document_id = file.name

        if file.suffix == ".pdf":
            paragraph_dicts = parse_pdf(file)
        elif file.suffix == ".txt":
            paragraph_dicts = parse_txt(file)
        else:
            raise ValueError(f"Unknown extension: {file.suffix}")

        if len(paragraph_dicts) < 3:
            log.warn(f"paper {document_id} only has {len(paragraph_dicts)} paragraphs")

        return Paper.parse_obj(
            dict(paragraphs=paragraph_dicts, document_id=document_id)
        )

    def sentences(self) -> Iterator[str]:
        for paragraph in self.paragraphs:
            for sentence in paragraph.sentences:
                yield sentence

    def nonempty_paragraphs(self) -> list[Paragraph]:
        return [p for p in self.paragraphs if not p.is_empty()]

    def __str__(self):
        return "\n\n".join(str(p) for p in self.paragraphs)

    def dict(self, *args, **kwargs):
        return super().dict(*args, **kwargs, exclude={"paragraphs"})



from collections.abc import Sequence
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import NewType
from typing import Protocol
from typing import runtime_checkable
from typing import TypeVar

from pydantic import BaseModel
from pydantic import root_validator
from pydantic import validator
from pydantic.generics import GenericModel

from ice.metrics.gold_standards import ParsedGoldStandardBase


T = TypeVar("T")

Stage = NewType("Stage", str)


class Sample(BaseModel):
    size: int | Literal["Unclear"] | None = None
    stage: Stage | None = None


class Arm(BaseModel):
    name: str
    description: str
    sample: Sample | None = None

    @root_validator(pre=True)
    def fix_keys(cls, values):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(values, dict) and "arm description" in values:
            values["description"] = values["arm description"]
        if isinstance(values, dict) and "initial sample" in values:
            values["sample"] = values["initial sample"]
        return values


class Experiment(BaseModel):
    name: str
    description: str
    arms: list[Arm]

    @root_validator(pre=True)
    def fix_keys(cls, values):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(values, dict) and "experiment description" in values:
            values["description"] = values["experiment description"]
        return values

    @validator("arms", pre=True)
    def fix_arms(cls, v):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(v, dict):
            return [Arm.parse_obj(vals | dict(name=k)) for k, vals in v.items()]
        return v


class ExperimentsArms(ParsedGoldStandardBase):
    question_short_name: ClassVar[str] = "experiments_arms"
    experiments: list[Experiment]

    @validator("experiments", pre=True)
    def validate_experiments(cls, v):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(v, dict):
            return [Experiment.parse_obj(vals | dict(name=k)) for k, vals in v.items()]
        return v


ReasoningStage = Literal["reasoning", "helpfulness", "answer"]


class PassageWithReasoning(GenericModel, Generic[T]):
    passage: Sequence[str]
    reasoning: str
    helpfulness: str | None = None
    score: float | None = None
    final_answer: T | None = None


@runtime_checkable
class MultipartReasoningPrompt(Protocol):
    def __call__(
        self,
        paragraphs: Sequence[str],
        helpfulness: str | None = None,
        reasoning: str | None = None,
    ) -> str:
        pass



import abc
import re
import sys

from contextvars import ContextVar
from io import TextIOWrapper
from typing import Any
from typing import cast
from typing import Optional
from typing import TypeVar

import nest_asyncio
import questionary
import rich

from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.pretty import pretty_repr
from rich.table import Table
from structlog import get_logger

log = get_logger()


class FloatZeroOneValidator(questionary.Validator):
    def validate(self, document):
        ok = re.match(r"^(0(\.\d+)?|1(\.0+)?)$", document.text)
        if not ok:
            raise questionary.ValidationError(
                message="Please enter a number between 0 and 1",
                cursor_position=len(document.text),
            )


class EnvironmentInterface(abc.ABC):
    def spinner(self, msg: str):
        raise NotImplementedError

    def print(
        self,
        msg: Any,
        *,
        format_markdown=False,
        wait_for_confirmation=False,
        file: None | str = None,
    ) -> None:
        raise NotImplementedError

    async def checkboxes(self, prompt: str, choices: list[str]) -> list[str]:
        raise NotImplementedError

    async def answer(self, prompt: str, default: str, multiline: bool = True) -> str:
        raise NotImplementedError

    async def score(
        self, question: str, context: str, default: Optional[float]
    ) -> float:
        # TODO: Make this generic classification instead of implicit single-class search
        raise NotImplementedError

    async def select(
        self, prompt: str, choices: list[str], default: Optional[str] = None
    ) -> str:
        raise NotImplementedError


Environment = TypeVar("Environment", bound=EnvironmentInterface)


class CliEnvironment(EnvironmentInterface):
    def __init__(self):
        try:
            nest_asyncio.apply()  # Needed for questionary to work
        except RuntimeError:
            log.warning("Nest_asyncio failed to apply.")
            pass
        self._console = Console()

    def spinner(self, msg):
        return self._console.status(msg)

    def print(
        self,
        msg: Any,
        *,
        format_markdown: bool = False,
        wait_for_confirmation: bool = False,
        file: str | None = None,
    ):
        if file:
            output = open(file, "a", encoding="utf-8")
        else:
            output = cast(TextIOWrapper, sys.stdout)

        # Create a console object with the appropriate file and width
        console = Console(file=output, width=80 if file else None)

        # Convert the message to a list of rich elements if needed
        if hasattr(msg, "to_rich_elements"):
            msg = msg.to_rich_elements()
        else:
            msg = [msg]

        # Print each element with the console object
        for element in msg:
            if isinstance(element, str):
                if format_markdown:
                    element = Markdown(element)
            elif isinstance(element, Table):
                pass
            else:
                element = pretty_repr(element)
            console.print(
                element, soft_wrap=True, crop=False, no_wrap=True, markup=False
            )

    def _maybe_print_prompt(self, prompt) -> str:
        if "\n" in prompt:
            rich.print(Panel(Markdown(prompt)))
            question = ""
        else:
            question = prompt
        return question

    async def checkboxes(self, prompt: str, choices: list[str]) -> list[str]:
        question = self._maybe_print_prompt(prompt)
        selections = questionary.checkbox(question, choices=choices).ask()
        if selections is None:
            print(prompt)
            raise ValueError("No selections provided")
        return cast(list[str], selections)

    async def answer(self, prompt: str, default: str, multiline: bool = True) -> str:
        question = self._maybe_print_prompt(prompt)
        answer = questionary.text(question, default=default, multiline=multiline).ask()
        if answer is None:
            print(prompt)
            raise ValueError("No answer provided")
        return cast(str, answer)

    async def select(
        self, prompt: str, choices: list[str], default: str | None = None
    ) -> str:
        question = self._maybe_print_prompt(prompt)
        answer = questionary.select(question, choices=choices, default=default).ask()
        if answer is None:
            print(prompt)
            raise ValueError("No answer provided")
        return cast(str, answer)

    async def score(
        self, question: str, context: str, default: Optional[float]
    ) -> float:
        score = questionary.text(
            f"Score the relevance of Context:\n{context}\n\nto Question: {question}",
            default=str(default or ""),
            validate=FloatZeroOneValidator,
        ).ask()
        if score is None:
            print(question)
            raise ValueError("No score provided")
        return float(cast(str, score))


_env = ContextVar("env", default=cast(EnvironmentInterface, CliEnvironment()))


def set_env(env: EnvironmentInterface):
    global _env
    _env.set(env)


def env() -> EnvironmentInterface:
    return _env.get()
